"""
Internal hook annotation, representation and calling machinery.
"""

import warnings
from collections.abc import Sequence
from typing import Any, Callable, List, Optional, Union

from .callers import HookCallError, HookResult, _multicall
from .exceptions import PluginCallError
from .implementation import HookImplementation, HookSpecification

HookExecFunc = Callable[
    ['HookCaller', List[HookImplementation], dict], HookResult
]
"""A function that loops calling a list of
:class:`~napari_plugin_engine.HookImplementation` s and returns a
:class:`~napari_plugin_engine.HookResult`.

Parameters
----------
hook_caller : HookCaller
    a :class:`HookCaller` instance.
hook_impls : List[HookImplementation]
    a list of :class:`~napari_plugin_engine.HookImplementation` instances to call.
kwargs : dict
    a mapping of keyword arguments to provide to the implementation.

Returns
-------
result : HookResult
    The :class:`~napari_plugin_engine.HookResult` object resulting from the call loop.
"""


class HookCaller:
    """The primary hook-calling object.

    A :class:`PluginManager` may have multiple ``HookCaller`` objects and they
    are stored in the ``plugin_manager.hook`` namespace, named after the `hook
    specification` that they represent.  For instance:

    .. code-block:: python

       pm = PluginManager("demo")
       pm.add_hookspec(some_module)
       # assuming `some_module` had an @hookspec named `my specification`
       assert isinstance(pm.hook.my_specification, HookCaller)

    Each ``HookCaller`` instance stores all of the :class:`HookImplementation` objects
    discovered during :meth:`plugin registration <PluginManager.register>`
    (each of which capture the implementation of a specific plugin for this
    hook specification).

    The ``HookCaller`` instance also usually creates and stores a reference to
    the :class:`HookSpecification` instance that encapsulates information about the hook
    specification, (at ``HookCaller.spec``)

    Parameters
    ----------
    name : str
        The name of the `hook specification` that this ``HookCaller``
        represents.
    hook_execute : Callable
        A :data:`.HookExecFunc` function.  In almost every case, this will be
        provided by the :class:`PluginManager` during hook registration as
        :meth:`PluginManager._hookexec`... which is, in turn, mostly just a
        wrapper around :func:`._multicall`.
    namespace : Any, optional
        An namespace (such as a module or class) to search during `HookSpecification`
        creation for functions decorated with ``@hookspec`` named with the
        string ``name``.
    spec_opts : Optional[dict], optional
        keyword arguments to be passed when creating the :class:`HookSpecification`
        instance at ``self.spec``.
    """

    def __init__(
        self,
        name: str,
        hook_execute: HookExecFunc,
        namespace: Any = None,
        spec_opts: Optional[dict] = None,
    ):
        self.name = name
        self._wrappers: List[HookImplementation] = []
        self._nonwrappers: List[HookImplementation] = []
        self._hookexec = hook_execute
        self.argnames = None
        self.kwargnames = None
        self.multicall = _multicall
        self.spec: Optional[HookSpecification] = None
        if namespace is not None:
            assert spec_opts is not None
            self.set_specification(namespace, spec_opts)

    def has_spec(self) -> bool:
        return self.spec is not None

    @property
    def is_firstresult(self) -> bool:
        return self.spec.firstresult if self.spec else False

    def set_specification(self, namespace, spec_opts):
        assert not self.has_spec()
        self.spec = HookSpecification(namespace, self.name, **spec_opts)
        if spec_opts.get("historic"):
            self._call_history = []

    def is_historic(self) -> bool:
        return hasattr(self, "_call_history")

    def _remove_plugin(self, plugin: Any):
        def remove(wrappers):
            for i, method in enumerate(wrappers):
                if method.plugin == plugin:
                    del wrappers[i]
                    return True

        if remove(self._wrappers) is None:
            if remove(self._nonwrappers) is None:
                raise ValueError("plugin %r not found" % (plugin,))

    def get_hookimpls(self) -> List[HookImplementation]:
        # Order is important for _hookexec
        return self._nonwrappers + self._wrappers

    def _add_hookimpl(self, hookimpl: HookImplementation):
        """Add an implementation to the callback chain."""
        if hookimpl.hookwrapper:
            methods = self._wrappers
        else:
            methods = self._nonwrappers

        if hookimpl.trylast:
            methods.insert(0, hookimpl)
        elif hookimpl.tryfirst:
            methods.append(hookimpl)
        else:
            # find last non-tryfirst method
            i = len(methods) - 1
            while i >= 0 and methods[i].tryfirst:
                i -= 1
            methods.insert(i + 1, hookimpl)

    def __repr__(self) -> str:
        return f"<HookCaller {self.name}>"

    def call_historic(
        self, result_callback=None, kwargs=None, with_impl=False
    ):
        """Call the hook with given ``kwargs`` for all registered plugins and
        for all plugins which will be registered afterwards.

        If ``result_callback`` is not ``None`` it will be called for for each
        non-``None`` result obtained from a hook implementation.

        If ``with_impl`` is ``True``, the caller is indicating that
        ``result_callback`` has a signature of ``callback(result, hookimpl)``,
        and will be called as such.
        """
        if result_callback is not None:
            result_callback._wants_impl = with_impl
        self._call_history.append((kwargs or {}, result_callback))
        # historizing hooks don't return results
        res = self._hookexec(self, self.get_hookimpls(), kwargs)
        if result_callback is None:
            return
        # XXX: remember firstresult isn't compat with historic
        if with_impl:
            for result, impl in zip(res.result, res.implementation):
                result_callback(result, impl)
        else:
            for x in res.result or []:
                result_callback(x)

    def call_extra(self, methods: List[Callable], kwargs: dict):
        """Call the hook with some additional temporarily participating
        methods using the specified ``kwargs`` as call parameters."""
        old = list(self._nonwrappers), list(self._wrappers)
        for method in methods:
            self._add_hookimpl(HookImplementation(method))
        try:
            return self(**kwargs)
        finally:
            self._nonwrappers, self._wrappers = old

    def _maybe_apply_history(self, method):
        """Apply call history to a new hookimpl if it is marked as historic."""
        if self.is_historic():
            for kwargs, result_callback in self._call_history:
                res = self._hookexec(self, [method], kwargs)
                if res.result and result_callback is not None:
                    if getattr(result_callback, '_wants_impl', False):
                        result_callback(res.result[0], res.implementation[0])
                    else:
                        result_callback(res.result[0])

    def get_plugin_implementation(self, plugin_name: str):
        """Return hook implementation instance for ``plugin_name`` if found."""
        try:
            return next(
                imp
                for imp in self.get_hookimpls()
                if imp.plugin_name == plugin_name
            )
        except StopIteration:
            raise KeyError(
                f"No implementation of {self.name!r} found "
                f"for plugin {plugin_name!r}."
            )

    def index(self, value: Union[str, HookImplementation]) -> int:
        """Return index of plugin_name or a HookImplementation in self._nonwrappers"""
        if isinstance(value, HookImplementation):
            return self._nonwrappers.index(value)
        elif isinstance(value, str):
            plugin_names = [imp.plugin_name for imp in self._nonwrappers]
            return plugin_names.index(value)
        else:
            raise TypeError(
                "argument provided to index must either be the "
                "(string) name of a plugin, or a HookImplementation instance"
            )

    def bring_to_front(
        self, new_order: Union[List[str], List[HookImplementation]]
    ):
        """Move items in ``new_order`` to the front of the call order.

        By default, hook implementations are called in last-in-first-out order
        of registration, and pluggy does not provide a built-in way to
        rearrange the call order of hook implementations.

        This function accepts a :class:`HookCaller` instance and the desired
        ``new_order`` of the hook implementations (in the form of list of
        plugin names, or a list of actual :class:`HookImplementation`
        instances) and reorders the implementations in the hook caller
        accordingly.

        .. note::

           Hook implementations are actually stored in *two* separate list
           attributes in the hook caller: :attr:`HookCaller._wrappers` and
           :attr:`HookCaller._nonwrappers`, according to whether the
           corresponding :class:`HookImplementation` instance was marked as a
           wrapper or not. This method *only* sorts _nonwrappers.

        Parameters
        ----------
        new_order :  list of str or list of :class:`HookImplementation`
            instances The desired CALL ORDER of the hook implementations.  The
            list does *not* need to include every hook implementation in
            :meth:`get_hookimpls`, but those that are not included will be left
            at the end of the call order.

        Raises
        ------
        TypeError
            If any item in ``new_order`` is neither a string (plugin_name) or a
            ``HookImplementation`` instance.
        ValueError
            If any item in ``new_order`` is neither the name of a plugin or a
            ``HookImplementation`` instance that is present in self._nonwrappers.
        ValueError
            If ``new_order`` argument has multiple entries for the same
            implementation.

        Examples
        --------
        Imagine you had a hook specification named ``print_plugin_name``, that
        expected plugins to simply print their own name. An implementation
        might look like:

        >>> # hook implementation for ``plugin_1``
        >>> @hook_implementation
        ... def print_plugin_name():
        ...     print("plugin_1")

        If three different plugins provided hook implementations. An example
        call for that hook might look like:

        >>> plugin_manager.hook.print_plugin_name()
        plugin_1
        plugin_2
        plugin_3

        If you wanted to rearrange their call order, you could do this:

        >>> new_order = ["plugin_2", "plugin_3", "plugin_1"]
        >>> plugin_manager.hook.print_plugin_name.bring_to_front(new_order)
        >>> plugin_manager.hook.print_plugin_name()
        plugin_2
        plugin_3
        plugin_1

        You can also just specify one or more item to move them to the front
        of the call order:
        >>> plugin_manager.hook.print_plugin_name.bring_to_front(["plugin_3"])
        >>> plugin_manager.hook.print_plugin_name()
        plugin_3
        plugin_2
        plugin_1
        """

        if not isinstance(new_order, Sequence) or isinstance(new_order, str):
            raise TypeError(
                'The first argument to "bring_to_front" '
                'must be a non-string sequence type.'
            )

        # make sure items in order are unique
        if len(new_order) != len(set(new_order)):
            raise ValueError("repeated item in order")

        # make new lists for the rearranged _nonwrappers
        # for details on the difference between wrappers and nonwrappers, see:
        # https://pluggy.readthedocs.io/en/latest/#wrappers
        _old_nonwrappers = self._nonwrappers.copy()
        _new_nonwrappers: List[HookImplementation] = []
        indices = [self.index(elem) for elem in new_order]
        for i in indices:
            # inserting because they get called in reverse order.
            _new_nonwrappers.insert(0, _old_nonwrappers[i])

        # remove items that have been pulled, leaving only items that
        # were not specified in ``new_order`` argument
        # do this rather than using .pop() above to avoid changing indices
        for i in sorted(indices, reverse=True):
            del _old_nonwrappers[i]

        # if there are any hook_implementations left over, add them to the
        # beginning of their respective lists (because at call time, these
        # lists are called in reverse order)
        if _old_nonwrappers:
            _new_nonwrappers = [x for x in _old_nonwrappers] + _new_nonwrappers

        # update the _nonwrappers list with the reordered list
        self._nonwrappers = _new_nonwrappers

    def _set_plugin_enabled(self, plugin_name: str, enabled: bool):
        """Enable or disable the hook implementation for a specific plugin.

        Parameters
        ----------
        plugin_name : str
            The name of a plugin implementing ``hook_spec``.
        enabled : bool
            Whether or not the implementation should be enabled.

        Raises
        ------
        KeyError
            If ``plugin_name`` has not provided a hook implementation for this
            hook specification.
        """
        self.get_plugin_implementation(plugin_name).enabled = enabled

    def enable_plugin(self, plugin_name: str):
        """enable implementation for ``plugin_name``."""
        self._set_plugin_enabled(plugin_name, True)

    def disable_plugin(self, plugin_name: str):
        """disable implementation for ``plugin_name``."""
        self._set_plugin_enabled(plugin_name, False)

    def _call_plugin(self, plugin_name: str, *args, **kwargs):
        """Call the hook implementation for a specific plugin

        .. note::

           This method is not intended to be called directly. Instead, just
           call the instance directly, specifing the ``_plugin`` argument.
           See the :meth:`__call__` method.

        Parameters
        ----------
        plugin_name : str
            Name of the plugin

        Returns
        -------
        Any
            Result of implementation call provided by plugin

        Raises
        ------
        TypeError
            If the implementation is a hook wrapper (cannot be called directly)
        TypeError
            If positional arguments are provided
        HookCallError
            If one of the required arguments in the hook specification is not
            present in ``kwargs``.
        PluginCallError
            If an exception is raised when calling the plugin
        """
        self._check_call_kwargs(kwargs)
        implementation = self.get_plugin_implementation(plugin_name)
        if implementation.hookwrapper:
            raise TypeError("Hook wrappers can not be called directly")

        # pluggy only allows calling hooks with keyword arguments
        if args:
            raise TypeError("hook calling supports only keyword arguments")
        _args: List[Any] = []
        # this converts kwargs into positional arguments in the correct order
        # for the hookspec
        try:
            _args = [kwargs[argname] for argname in implementation.argnames]
        except KeyError:
            for argname in implementation.argnames:
                if argname not in kwargs:
                    raise HookCallError(
                        f"hook call must provide argument {argname}"
                    )

        try:
            return implementation(*_args)
        except Exception as exc:
            raise PluginCallError(implementation) from exc

    def call_with_result_obj(
        self, *, _skip_impls: List[HookImplementation] = list(), **kwargs
    ) -> HookResult:
        """Call hook implementation(s) for this spec and return HookResult.

        The :class:`HookResult` object carries the result (in its ``result``
        property) but also additional information about the hook call, such
        as the implementation that returned each result and any call errors.

        Parameters
        ----------
        _skip_impls : List[HookImplementation], optional
            A list of HookImplementation instances that should be *skipped* when calling
            hook implementations, by default None
        **kwargs
            keys should match the names of arguments in the corresponding hook
            specification, values will be passed as arguments to the hook
            implementations.

        Returns
        -------
        result : HookResult
            A :class:`HookResult` object that contains the results returned by
            plugins along with other metadata about the call.

        Raises
        ------
        HookCallError
            If one or more of the keys in ``kwargs`` is not present in
            one of the ``hook_impl.argnames``.
        PluginCallError
            If ``firstresult == True`` and a plugin raises an Exception.
        """
        # if not self.get_hookimpls():
        #     warnings.warn(
        #         'No hook implementations registered for this hook caller!'
        #     )
        self._check_call_kwargs(kwargs)
        impls = [imp for imp in self.get_hookimpls() if imp not in _skip_impls]
        return self._hookexec(self, impls, kwargs)

    def __call__(
        self,
        *args,
        _plugin: Optional[str] = None,
        _skip_impls: List[HookImplementation] = list(),
        **kwargs,
    ) -> Union[Any, List[Any]]:
        """Call hook implementation(s) for this spec and return result(s).

        This is the primary way to call plugin hook implementations.

        .. note::

           Parameters are prefaced by underscores to reduce potential conflicts
           with argument names in hook specifications.  There is a test in
           :func:`test_hook_specifications.test_annotation_on_hook_specification`
           to ensure that these argument names are never used in one of our
           hookspecs.

        Parameters
        ----------
        _plugin : str, optional
            The name of a specific plugin to use.  By default all
            implementations will be called (though if ``firstresult==True``,
            only the first non-None result will be returned).
        _skip_impls : List[HookImplementation], optional
            A list of HookImplementation instances that should be *skipped* when calling
            hook implementations, by default None
        **kwargs
            keys should match the names of arguments in the corresponding hook
            specification, values will be passed as arguments to the hook
            implementations.

        Raises
        ------
        HookCallError
            If one or more of the keys in ``kwargs`` is not present in one of
            the ``hook_impl.argnames``.
        PluginCallError
            If ``firstresult == True`` and a plugin raises an Exception.

        Returns
        -------
        result
            If the hookspec was declared with ``firstresult==True``, a single
            result will be returned. Otherwise will return a list of results
            from all hook implementations for this hook caller.

            If ``_plugin`` is provided, will return the single result from the
            specified plugin.
        """
        if args:
            raise TypeError("hook calling supports only keyword arguments")
        if _plugin:
            # if a plugin name is specified, just call it directly
            return self._call_plugin(_plugin, **kwargs)

        result = self.call_with_result_obj(_skip_impls=_skip_impls, **kwargs)
        return result.result

    def _check_call_kwargs(self, kwargs):
        """Warn if any keys in the hookspec are not present in this call.

        It's possible to add arguments to hook specifications (as they evolve).
        Here we just emit a warning if there are arguments in the hookspec that
        were not specified in this call, which may mean the call could be
        updated.
        """
        # "historic" hooks can be called with ``call_historic()`` *before*
        # having been registered.  However they must be called with
        # self.call_historic().
        # https://pluggy.readthedocs.io/en/latest/index.html#historic-hooks
        assert (
            not self.is_historic()
        ), 'Historic hooks must be called with `call_historic()`'

        if self.spec and self.spec.argnames:
            notincall = set(self.spec.argnames) - set(kwargs.keys())
            if notincall:
                warnings.warn(
                    "Argument(s) {} which are declared in the hookspec "
                    "can not be found in this hook call".format(
                        tuple(notincall)
                    ),
                    stacklevel=2,
                )



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

sys.path.insert(0, os.path.abspath('..'))


# -- Project information -----------------------------------------------------

project = 'napari-plugin-engine'
author = 'Holger Krekel & Talley Lambert'
master_doc = 'index'
html_show_sourcelink = False

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.todo',
    'sphinx.ext.viewcode',
    'sphinx.ext.autodoc',
    'sphinx.ext.napoleon',
    'sphinx_autodoc_typehints',
    'sphinx.ext.autosummary',
    'sphinx.ext.intersphinx',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']
pygments_style = 'paraiso-dark'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# this is for "press" theme
# html_css_files = ['css/press.css']
html_css_files = ['css/rtd.css']

intersphinx_mapping = {
    'python': ('https://docs.python.org/3', None),
}



from contextlib import contextmanager

import pytest

from napari_plugin_engine import (
    HookCaller,
    HookImplementation,
    HookImplementationMarker,
    HookSpecification,
    HookSpecificationMarker,
    PluginManager,
)


@pytest.fixture
def test_plugin_manager() -> PluginManager:
    """A plugin manager fixture with the project name 'test'."""
    return PluginManager(project_name='test')


@pytest.fixture
def add_specification(test_plugin_manager):
    """Return a decorator that adds a HookSpecification to test_plugin_manager."""

    def addspec(function=None, *, firstresult=False, historic=False):
        def wrap(func):
            project = test_plugin_manager.project_name
            test_hookspec = HookSpecificationMarker(project)
            test_hookspec(firstresult=firstresult, historic=historic)(func)
            name = func.__name__
            namespace = type("Hook", (), {name: func})
            assert not hasattr(
                test_plugin_manager.hook, name
            ), f"Hook already exists with name: {name}"
            opts = getattr(func, HookSpecification.format_tag(project))
            hook_caller = HookCaller(
                name, test_plugin_manager._hookexec, namespace, opts
            )
            setattr(test_plugin_manager.hook, name, hook_caller)

        return wrap(function) if function is not None else wrap

    return addspec


@pytest.fixture
def add_implementation(test_plugin_manager):
    """Return a decorator that adds a HookImplementation to test_plugin_manager."""

    def addimpl(
        function=None,
        *,
        specname=None,
        tryfirst=False,
        trylast=False,
        hookwrapper=False,
    ):
        def wrap(func):
            project = test_plugin_manager.project_name
            HookImplementationMarker(project)(
                tryfirst=tryfirst,
                trylast=trylast,
                hookwrapper=hookwrapper,
                specname=specname,
            )(func)
            _specname = specname or func.__name__
            hook_caller = getattr(test_plugin_manager.hook, _specname, None)
            assert hook_caller, f"No hook with with name: {_specname}"
            opts = getattr(func, HookImplementation.format_tag(project))
            hook_caller._add_hookimpl(HookImplementation(func, **opts))
            return func

        return wrap(function) if function is not None else wrap

    return addimpl


@pytest.fixture
def caller_from_implementation(
    test_plugin_manager, add_specification, add_implementation
):
    """Return hook caller with implementation as its own spec definition.

    Adds a specification and implementation to the test_plugin_manager based on
    a single function definition (e.g. assumes that the implementation has the
    correct signature).  Returns the hook caller instance prepopulated with the
    hook implementation.
    """

    def wrap(func, spec_kwargs={}, impl_kwargs={}):
        add_specification(func, **spec_kwargs)
        add_implementation(func, **impl_kwargs)
        name = spec_kwargs.get('specname') or func.__name__
        return getattr(test_plugin_manager.hook, name)

    return wrap


@pytest.fixture
def temporary_hookimpl(test_plugin_manager):
    """A fixture that can be used to insert a HookImplementation in the hook call loop.

    Use as a context manager, which will return the hook_caller for the
    corresponding hook specification.

    Example
    -------
    .. code-block: python

        def my_hook_implementation(arg):
            raise ValueError("shoot!")

        with temporary_hookimpl(my_hook_implementation) as hook_caller:
            with pytest.raises(PluginCallError):
                hook_caller(arg=42)
    """

    @contextmanager
    def wrap(func, specname=None, *, tryfirst=True, trylast=None):
        project = test_plugin_manager.project_name
        marker = HookImplementationMarker(project)
        marker(tryfirst=tryfirst, trylast=trylast, specname=specname)(func)
        _specname = specname or func.__name__
        hook_caller = getattr(test_plugin_manager.hook, _specname, None)
        assert hook_caller, f"No hook with with name: {_specname}"
        opts = getattr(func, HookImplementation.format_tag(project))
        impl = HookImplementation(func, **opts)
        hook_caller._add_hookimpl(impl)
        try:
            yield hook_caller
        finally:
            if impl in hook_caller._nonwrappers:
                hook_caller._nonwrappers.remove(impl)
            if impl in hook_caller._wrappers:
                hook_caller._wrappers.remove(impl)
            assert impl not in hook_caller.get_hookimpls()

    return wrap



import functools
import importlib
import inspect
import os
import re
import sys
import warnings
from contextlib import contextmanager
from logging import getLogger
from pathlib import Path
from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    List,
    Optional,
    Set,
    Tuple,
    Type,
    Union,
)

from . import _tracing
from .callers import HookResult
from .dist import (
    _top_level_module_to_dist,
    get_metadata,
    importlib_metadata,
    standard_metadata,
)
from .exceptions import (
    Empty,
    PluginError,
    PluginImportError,
    PluginRegistrationError,
    PluginValidationError,
    _empty,
)
from .hooks import HookCaller, HookExecFunc
from .implementation import HookImplementation, HookSpecification
from .markers import HookImplementationMarker, HookSpecificationMarker

logger = getLogger(__name__)


class PluginManager:
    """Core class which manages registration of plugin objects and hook calls.

    You can register new hooks by calling :meth:`~PluginManager.add_hookspecs`.
    You can register plugin objects (which contain hooks) by calling
    :meth:`~PluginManager.register`.  The ``PluginManager`` is initialized with
    a ``project_name`` that is used when discovering *hook specifications* and
    *hook implementations*.

    For debugging purposes you may call :meth:`.PluginManager.enable_tracing`
    which will subsequently send debug information to the trace helper.

    Parameters
    ----------
    project_name : str
        The name of the host project.  All :class:`HookImplementationMarker`
        and :class:`HookSpecificationMarker` instances must be created using
        the same ``project_name`` to be detected by this plugin manager.
    discover_entry_point : str, optional
        The default entry_point group to search when discovering plugins with
        :meth:`PluginManager.discover`, by default None
    discover_prefix : str, optional
        The default module prefix to use when discovering plugins with
        :meth:`PluginManager.discover`, by default None
    discover_path : str or list of str, optional
        A path or paths to include when discovering plugins with
        :meth:`PluginManager.discover`, by default None

    Examples
    --------

    .. code-block:: python

        from napari_plugin_engine import PluginManager
        import my_hookspecs

        plugin_manager = PluginManager(
            'my_project',
            discover_entry_point='app.plugin',
            discover_prefix='app_',
        )
        plugin_manager.add_hookspecs(my_hookspecs)
        plugin_manager.discover()

        # hooks now live in plugin_manager.hook
        # plugin dict is at plugin_manager.plugins
    """

    def __init__(
        self,
        project_name: str,
        *,
        discover_entry_point: Optional[str] = None,
        discover_prefix: Optional[str] = None,
        discover_path: Optional[List[str]] = None,
    ):
        self.project_name = project_name
        self.discover_entry_point = discover_entry_point
        self.discover_prefix = discover_prefix
        self.discover_path = discover_path or []
        #: dict : mapping of ``plugin_name`` → ``plugin`` (object)
        #:
        #: Plugins get added to this dict in :meth:`~PluginManager.register`
        self.plugins: Dict[str, Any] = {}
        #: dict : mapping of ``plugin`` (object) → list of :class:`HookCaller`
        #:
        #: :class:`HookCaller` s get added in :meth:`~PluginManager.register`
        self._plugin2hookcallers: Dict[Any, List[HookCaller]] = {}

        self._blocked: Set[str] = set()
        # multiple plugins might register the same entry point
        # _id_counts tracks the count of each identical entry point
        self._id_counts: Dict[str, int] = {}

        self.trace = _tracing.TagTracer().get("pluginmanage")
        self.hook = _HookRelay(self)
        self._inner_hookexec: HookExecFunc = lambda c, m, k: c.multicall(
            m, k, firstresult=c.is_firstresult
        )

    @property
    def hooks(self) -> '_HookRelay':
        """An alias for PluginManager.hook"""
        return self.hook

    def _hookexec(
        self,
        caller: HookCaller,
        methods: List[HookImplementation],
        kwargs: dict,
    ) -> HookResult:
        """Returns a function that will call a set of hookipmls with a caller.

        This function will be passed to ``HookCaller`` instances that are
        created during hookspec and plugin registration.

        If :meth:`~.PluginManager.enable_tracing` is used, it will set it's own
        wrapper function at self._inner_hookexec to enable tracing of hook
        calls.

        Parameters
        ----------
        caller : HookCaller
            The HookCaller instance that will call the HookImplementations.
        methods : List[HookImplementation]
            A list of :class:`~napari_plugin_engine.HookImplementation` objects
            whose functions will be called during the hook call loop.
        kwargs : dict
            Keyword arguments to pass when calling the ``HookImplementation``.

        Returns
        -------
        :class:`~napari_plugin_engine.HookResult`
            The result object produced by the multicall loop.
        """
        return self._inner_hookexec(caller, methods, kwargs)

    def iter_available(
        self,
        path: Optional[str] = None,
        entry_point: Optional[str] = None,
        prefix: Optional[str] = None,
    ) -> Generator[Tuple[str, str, Optional[str]], None, None]:
        """Iterate over available plugins.

        Parameters
        ----------
        path : str, optional
            If a string is provided, it is added to ``sys.path`` (and
            ``self.discover_path``) before importing, and removed at the end.
        entry_point : str, optional
            An entry_point group to search for, by default
            ``self.discover_entry_point`` is used
        prefix : str, optional
            If ``provided``, modules in the environment starting with
            ``prefix`` will be imported and searched for hook implementations
            by default ``self.discover_prefix`` is used

        Notes
        -----
        See docstring of :func:`iter_available_plugins` for details.

        """
        _path = self.discover_path
        if path:
            _path.append(path)
        yield from iter_available_plugins(
            entry_point or self.discover_entry_point,
            prefix or self.discover_prefix,
            _path,
            include_uninstalled=bool(self.discover_prefix),
        )

    def discover(
        self,
        path: Optional[str] = None,
        entry_point: Optional[str] = None,
        prefix: Optional[str] = None,
        ignore_errors: bool = True,
    ) -> Tuple[int, List[PluginError]]:
        """Discover and load plugins.

        Parameters
        ----------
        path : str, optional
            If a string is provided, it is added to ``sys.path`` (and
            ``self.discover_path``) before importing, and removed at the end.
        entry_point : str, optional
            An entry_point group to search for, by default
            ``self.discover_entry_point`` is used
        prefix : str, optional
            If ``provided``, modules in the environment starting with
            ``prefix`` will be imported and searched for hook implementations
            by default ``self.discover_prefix`` is used
        ignore_errors : bool, optional
            If ``True``, errors will be gathered and returned at the end.
            Otherwise, they will be raised immediately. by default True

        Returns
        -------
        (count, errs) : Tuple[int, List[PluginError]]
            The number of succefully loaded modules, and a list of errors that
            occurred (if ``ignore_errors`` was ``True``)
        """
        self.hook._needs_discovery = False
        # allow debugging escape hatch
        if os.environ.get("DISABLE_ALL_PLUGINS"):
            warnings.warn(
                'Plugin discovery disabled due to '
                'environmental variable "DISABLE_ALL_PLUGINS"'
            )
            return 0, []

        errs: List[PluginError] = []
        count = 0
        for name, mod_name, dist_name in self.iter_available(
            path, entry_point, prefix
        ):
            old_name = name
            # different plugin has already registered this entry point
            if self.is_registered(name):
                mod_names = (
                    plugin_mod.__name__ for plugin_mod in self.plugins.values()
                )
                # we may have registered this entry point under a different name,
                # so check module names to avoid duplicate registration
                if mod_name in mod_names:
                    continue
                new_name = f"{name}-{self._id_counts[name]}"
                previously_registered_mod = self.plugins[name].__name__
                warnings.warn(
                    f"Plugin {name} already registered by module "
                    + f"{previously_registered_mod}! Registering as {new_name}."
                )
                name = new_name
            elif self.is_blocked(name):
                continue

            try:
                if self._load_and_register(mod_name, name):
                    count += 1
                    self._id_counts[name] = 1
                    self._id_counts[old_name] += 1
            except PluginError as e:
                errs.append(e)
                # commenting out for now, because napari stores this blockage
                # too permanently, and it's hard to differentiate between
                # plugins intentionally blocked by the user.
                # self.set_blocked(name)
                if ignore_errors:
                    continue
                raise e

        return count, errs

    @contextmanager
    def discovery_blocked(self) -> Generator:
        """A context manager that temporarily blocks discovery of new plugins."""
        current = self.hook._needs_discovery
        self.hook._needs_discovery = False
        try:
            yield
        finally:
            self.hook._needs_discovery = current

    def _load_and_register(
        self, mod_name: str, plugin_name: Optional[str] = None
    ) -> Optional[str]:
        """A helper function to import and register a module as ``plugin_name``.

        Parameters
        ----------
        mod : str
            The name of a module (or class in a module) to load.
        plugin_name : str, optional
            Optional name for plugin, by default ``get_canonical_name(plugin)``

        Returns
        -------
        str or None
            canonical plugin name, or ``None`` if the name is blocked from
            registering.

        Raises
        ------
        PluginImportError
            If an exception is raised when importing the module.
        PluginRegistrationError
            If an entry_point is declared that is neither a module nor a class.
        PluginRegistrationError
            If an exception is raised during plugin registration.
        """
        try:
            module = load(mod_name)
            if self.is_registered(module):
                return None
        except Exception as exc:
            raise PluginImportError(
                f'Error while importing module {mod_name}',
                plugin_name=plugin_name,
                cause=exc,
            )
        if not (inspect.isclass(module) or inspect.ismodule(module)):
            raise PluginRegistrationError(
                f'Plugin "{plugin_name}" declared entry_point "{mod_name}"'
                ' which is neither a module nor a class.',
                plugin=module,
                plugin_name=plugin_name,
            )

        try:
            return self.register(module, plugin_name)
        except PluginError:
            raise
        except Exception as exc:
            raise PluginRegistrationError(
                plugin=module, plugin_name=plugin_name, cause=exc
            )

    def register(
        self, namespace: Any, name: Optional[str] = None
    ) -> Optional[str]:
        """Register a plugin and return its canonical name or ``None``.

        Parameters
        ----------
        namespace : Any
            The namespace (class, module, dict, etc...) of the plugin to register
        name : str, optional
            Optional name for plugin, by default ``get_canonical_name(plugin)``

        Returns
        -------
        str or None
            canonical plugin name, or ``None`` if the name is blocked from
            registering.

        Raises
        ------
        TypeError
            If ``namespace`` is a string.
        ValueError
            if the plugin ``name`` or ``namespace`` is already registered.
        """
        if isinstance(namespace, str):
            raise TypeError("Plugin objects cannot be strings.")

        if isinstance(namespace, dict):
            return self._register_dict(namespace, name)

        plugin_name = name or get_canonical_name(namespace)

        if self.is_blocked(plugin_name):
            return None

        if self.is_registered(plugin_name):
            raise ValueError(f"Plugin name already registered: {plugin_name}")
        if self.is_registered(namespace):
            raise ValueError(f"Plugin module already registered: {namespace}")

        hookcallers = []
        for hookimpl in iter_implementations(namespace, self.project_name):
            hookimpl.plugin_name = plugin_name
            hook_caller = getattr(self.hook, hookimpl.specname, None)
            # if we don't yet have a hookcaller by this name, create one.
            if hook_caller is None:
                hook_caller = HookCaller(hookimpl.specname, self._hookexec)
                setattr(self.hook, hookimpl.specname, hook_caller)
            # otherwise, if it has a specification, validate the new
            # hookimpl against the specification.
            elif hook_caller.has_spec():
                self._verify_hook(hook_caller, hookimpl)
                hook_caller._maybe_apply_history(hookimpl)
            # Finally, add the hookimpl to the hook_caller and the hook
            # caller to the list of callers for this plugin.
            hook_caller._add_hookimpl(hookimpl)
            hookcallers.append(hook_caller)

        self._plugin2hookcallers[namespace] = hookcallers
        if not hookcallers:
            ns_name = get_canonical_name(namespace)
            warnings.warn(
                f"Module {ns_name!r} from plugin {plugin_name!r} has no hooks! "
                "Consider disabling or uninstalling this plugin. "
                "If you are a developer, please check your entry point."
            )
        self.plugins[plugin_name] = namespace
        return plugin_name

    def _register_dict(
        self, dct: Dict[str, Callable], name: Optional[str] = None, **kwargs
    ) -> Optional[str]:
        """Register a dict as a mapping of method name -> method.

        Parameters
        ----------
        dct : Dict[str, Callable]
            Mapping of method name to method.
        name : Optional[str], optional
            The plugin_name to assign to this object, by default None

        Returns
        -------
        str or None
            canonical plugin name, or ``None`` if the name is blocked from
            registering.
        """
        mark = HookImplementationMarker(self.project_name)
        clean_dct = {
            key: mark(specname=key, **kwargs)(val)
            for key, val in dct.items()
            if inspect.isfunction(val)
        }
        namespace = ensure_namespace(clean_dct)
        return self.register(namespace, name)

    def get_name(self, plugin):
        """Return name for registered plugin or ``None`` if not registered."""
        for name, val in self.plugins.items():
            if plugin == val:
                return name

    def _ensure_plugin(self, name_or_object: Any) -> Any:
        """Return plugin object given a name or object. Or raise an exception.

        Parameters
        ----------
        name_or_object : Any
            Either a string (in which case it is interpreted as a plugin name),
            or a non-string object (in which case it is assumed to be a plugin
            module or class).

        Returns
        -------
        Any
            The plugin object, if found.

        Raises
        ------
        KeyError
            If the plugin does not exist.
        """
        if isinstance(name_or_object, str):
            plugin_name = name_or_object
        else:
            plugin_name = self.get_name(name_or_object)

        if plugin_name in self.plugins:
            return self.plugins[plugin_name]

        if isinstance(name_or_object, str):
            msg = f"No plugin found with the name {name_or_object}"
        else:
            msg = f"No plugin found with the name {name_or_object}"
        raise KeyError(msg)

    def unregister(self, name_or_object: Any) -> Optional[Any]:
        """Unregister a plugin object or ``plugin_name``.

        Parameters
        ----------
        name_or_object : str or Any
            A module/class object or a plugin name (string).

        Returns
        -------
        module : Any or None
            The module object, or None if the ``name_or_object`` was not found.
        """
        try:
            plugin = self._ensure_plugin(name_or_object)
        except KeyError as e:
            warnings.warn(str(e))
            return None

        del self.plugins[self.get_name(plugin)]

        for hookcaller in self._plugin2hookcallers.pop(plugin, []):
            hookcaller._remove_plugin(plugin)

        return plugin

    def prune(self):
        """Unregister modules that can no longer be imported.

        Useful if pip uninstall has been run during the session.
        """
        _top_level_module_to_dist.cache_clear()
        for plugin_module in list(self.plugins.values()):
            try:
                importlib.reload(plugin_module)
            except ModuleNotFoundError:
                self.unregister(plugin_module)

    def _add_hookspec_dict(self, dct: Dict[str, Callable], **kwargs):
        mark = HookSpecificationMarker(self.project_name)
        clean_dct = {
            key: mark(**kwargs)(val)
            for key, val in dct.items()
            if inspect.isfunction(val)
        }
        namespace = ensure_namespace(clean_dct)
        return self.add_hookspecs(namespace)

    def add_hookspecs(self, namespace: Any):
        """Add new hook specifications defined in the given ``namespace``.

        Functions are recognized if they have been decorated accordingly.
        """

        if isinstance(namespace, dict):
            return self._add_hookspec_dict(namespace)

        names = []
        for name in dir(namespace):
            method = getattr(namespace, name)
            if not inspect.isroutine(method):
                continue

            tag = HookSpecification.format_tag(self.project_name)
            spec_opts = getattr(method, tag, None)
            if spec_opts is not None:
                hook_caller = getattr(self.hook, name, None)
                if hook_caller is None:
                    hook_caller = HookCaller(
                        name, self._hookexec, namespace, spec_opts
                    )
                    setattr(self.hook, name, hook_caller)
                else:
                    # plugins registered this hook without knowing the spec
                    hook_caller.set_specification(namespace, spec_opts)
                    for hookfunction in hook_caller.get_hookimpls():
                        self._verify_hook(hook_caller, hookfunction)
                names.append(name)

        if not names:
            raise ValueError(
                f"did not find any {self.project_name!r} hooks in {namespace!r}"
            )

    def is_registered(self, obj: Any) -> bool:
        """Return ``True`` if the plugin is already registered."""
        if isinstance(obj, str):
            return obj in self.plugins
        return obj in self._plugin2hookcallers

    def is_blocked(self, plugin_name: str) -> bool:
        """Return ``True`` if the given plugin name is blocked."""
        return plugin_name in self._blocked

    def set_blocked(self, plugin_name: str, blocked=True):
        """Block registrations of ``plugin_name``, unregister if registered.

        Parameters
        ----------
        plugin_name : str
            A plugin name to block.
        blocked : bool, optional
            Whether to block the plugin.  If ``False`` will "unblock"
            ``plugin_name``.  by default True
        """
        if blocked:
            self._blocked.add(plugin_name)
            if self.is_registered(plugin_name):
                self.unregister(plugin_name)
        else:
            if plugin_name in self._blocked:
                self._blocked.remove(plugin_name)

    def get_errors(
        self,
        plugin: Union[Any, Empty] = _empty,
        error_type: Union[Type['PluginError'], Empty] = _empty,
    ) -> List[PluginError]:
        """Return a list of PluginErrors associated with ``plugin``.

        Parameters
        ----------
        plugin : Any
            If provided, will restrict errors to those that were raised by
            ``plugin``.  If a string is provided, it will be interpreted as the
            name of the plugin, otherwise it is assumed to be the actual plugin
            object itself.
        error_type : PluginError
            If provided, will restrict errors to instances of ``error_type``.
        """
        # not using _ensure_plugin because it may not have been successfully
        # registered
        plugin_name: Union[str, Empty] = _empty
        if isinstance(plugin, str):
            plugin_name = plugin
            plugin = _empty
        return PluginError.get(
            plugin=plugin, plugin_name=plugin_name, error_type=error_type
        )

    def _verify_hook(
        self, hook_caller: HookCaller, hookimpl: HookImplementation
    ):
        """Check validity of a ``hookimpl``

        Parameters
        ----------
        hook_caller : HookCaller
            A :class:`HookCaller` instance.
        hookimpl : HookImplementation
            A :class:`HookImplementation` instance, implementing the hook in
            ``hook_caller``.

        Raises
        ------
        PluginValidationError
            If hook_caller is historic and the hookimpl is a hookwrapper.
        PluginValidationError
            If there are any argument names in the ``hookimpl`` that are not
            in the ``hook_caller.spec``.

        Warns
        -----
        Warning
            If the hookspec has ``warn_on_impl`` flag (usually a deprecation).
        """
        # historic hooks cannot have hookwrappers
        if hook_caller.is_historic() and hookimpl.hookwrapper:
            raise PluginValidationError(
                hookimpl,
                f"Plugin {hookimpl.plugin_name!r}\nhook "
                f"{hook_caller.name!r}\nhistoric incompatible to hookwrapper",
            )

        if not hook_caller.spec:
            return

        # If the hookspec has ``warn_on_impl`` flag show a warning.
        if hook_caller.spec.warn_on_impl:
            warnings.warn_explicit(
                hook_caller.spec.warn_on_impl,
                type(hook_caller.spec.warn_on_impl),
                lineno=hookimpl.function.__code__.co_firstlineno,
                filename=hookimpl.function.__code__.co_filename,
            )

        # If there are any argument names in the hookimpl that are not
        # in the hook specification.
        notinspec = set(hookimpl.argnames) - set(hook_caller.spec.argnames)
        if notinspec:
            raise PluginValidationError(
                hookimpl,
                f"Plugin {hookimpl.plugin_name!r} for hook {hook_caller.name!r}"
                f"\nhookimpl definition: {_formatdef(hookimpl.function)}\n"
                f"Argument(s) {notinspec} are declared in the hookimpl but "
                "can not be found in the hookspec",
            )

    def check_pending(self):
        """Make sure all hooks have a specification, or are optional.

        Raises
        ------
        PluginValidationError
            If a hook implementation that was *not* marked as ``optionalhook``
            has been registered for a non-existent hook specification.
        """
        for name in self.hook.__dict__:
            if name.startswith("_"):
                continue
            hook = getattr(self.hook, name)
            if not hook.has_spec():
                for hookimpl in hook.get_hookimpls():
                    if not hookimpl.optionalhook:
                        raise PluginValidationError(
                            hookimpl,
                            f"unknown hook {name!r} in "
                            f"plugin {hookimpl.plugin!r}",
                        )

    def get_hookcallers(self, plugin: Any) -> Optional[List[HookCaller]]:
        """get all hook callers for the specified plugin."""
        return self._plugin2hookcallers.get(plugin)

    def add_hookcall_monitoring(
        self,
        before: Callable[[str, List[HookImplementation], dict], None],
        after: Callable[
            [HookResult, str, List[HookImplementation], dict], None
        ],
    ) -> Callable[[], None]:
        """Add before/after tracing functions for all hooks.

        return an undo function which, when called, will remove the added
        tracers.

        ``before(hook_name, hook_impls, kwargs)`` will be called ahead of all
        hook calls and receive a hookcaller instance, a list of HookImplementation
        instances and the keyword arguments for the hook call.

        ``after(outcome, hook_name, hook_impls, kwargs)`` receives the same
        arguments as ``before`` but also a
        :py:class:`napari_plugin_engine.callers._Result` object which
        represents the result of the overall hook call.
        """
        oldcall = self._inner_hookexec

        def traced_hookexec(
            caller: HookCaller, impls: List[HookImplementation], kwargs: dict
        ):
            before(caller.name, impls, kwargs)
            outcome = HookResult.from_call(
                lambda: oldcall(caller, impls, kwargs)
            )
            after(outcome, caller.name, impls, kwargs)
            return outcome

        self._inner_hookexec = traced_hookexec

        def undo():
            self._inner_hookexec = oldcall

        return undo

    def enable_tracing(self):
        """Enable tracing of hook calls and return an undo function."""
        hooktrace = self.trace.root.get("hook")

        def before(hook_name, methods, kwargs):
            hooktrace.root.indent += 1
            hooktrace(hook_name, kwargs)

        def after(outcome, hook_name, methods, kwargs):
            if outcome.excinfo is None:
                hooktrace("finish", hook_name, "-->", outcome.result)
            hooktrace.root.indent -= 1

        return self.add_hookcall_monitoring(before, after)

    def get_metadata(
        self, plugin: Any, *values
    ) -> Optional[Union[str, Dict[str, Optional[str]]]]:
        """Return metadata values for a given plugin

        Parameters
        ----------
        plugin : Any
            Either a string (in which case it is interpreted as a plugin name),
            or a non-string object (in which case it is assumed to be a plugin
            module or class).
        *values : str
            key(s) to lookup in the plugin object distribution metadata.  At
            least one value must be supplied.

        Raises
        ------
        TypeError
            If no values are supplied.
        KeyError
            If the plugin does not exist.
        """
        if not values:
            raise TypeError(
                'get_metadata() requires at least one positional '
                'argument: the metadata value(s) to lookup'
            )
        # allow other objects to pass through directly to get_metadata
        if isinstance(plugin, str):
            plugin = self._ensure_plugin(plugin)
        return get_metadata(plugin, *values)

    def get_standard_metadata(self, plugin: Any):
        """Return a standard metadata dict for ``plugin``.

        Parameters
        ----------
        plugin : Any
            A plugin name or any object.  If it is a plugin name, it
            *must* be a registered plugin.

        Returns
        -------
        metadata : dict
            A  dicts with plugin metadata. The dict is guaranteed to have the
            following keys:

            - **plugin_name**: The name of the plugin as registered
            - **package**: The name of the package
            - **version**: The version of the plugin package
            - **summary**: A one-line summary of what the distribution does
            - **author**: The author’s name
            - **email**: The author’s (or maintainer's) e-mail address.
            - **license**: The license covering the distribution
            - **url**: The home page for the package, or dowload url if N/A.
            - **hooks**: A list of hookspec names that this plugin implements.

        Raises
        ------
        KeyError
            If ``plugin`` is a string, but is not a registered plugin_name.
        """
        if isinstance(plugin, str):
            plugin = self._ensure_plugin(plugin)
        plugin_name = self.get_name(plugin)
        # TODO: is there no better API for this already?
        hooks = []
        for caller in self._plugin2hookcallers[plugin]:
            try:
                impl = caller.get_plugin_implementation(plugin_name)
                hooks.append(impl.specname)
            except KeyError:
                pass
        plugin_meta = dict(plugin_name=plugin_name, hooks=hooks)
        plugin_meta.update(standard_metadata(plugin))
        return plugin_meta

    def list_plugin_metadata(self) -> List[Dict[str, Optional[str]]]:
        """Return list of standard metadata dicts for every registered plugin.

        Returns
        -------
        metadata : dict
            A list of dicts with plugin metadata. Every dict in the list is
            guaranteed to have the following keys mentioned in
            :meth:`~PluginManager.get_standard_metadata`
        """
        return [
            self.get_standard_metadata(plugin)
            for plugin in self._plugin2hookcallers
        ]

    def __str__(self) -> str:
        nhooks = len(self.hooks)
        nplug = len(self.plugins)
        text = f'PluginManager for "{self.project_name}"\n'
        text += f'({nhooks} hook specs and {nplug} plugins)\n'
        text += '-' * 45 + '\n'
        for name, plugin in sorted(self.plugins.items(), key=lambda x: x[0]):
            text += self.plugin_info(plugin) + "\n"

        if self._blocked:
            text += '\nBlocked Plugins:\n----------------\n'
            text += "\n".join(self._blocked)
        return text

    def plugin_info(self, plugin) -> str:
        plugin = self._ensure_plugin(plugin)
        plugin_name = self.get_name(plugin)
        version = self.get_metadata(plugin, 'version')
        hooks = self._plugin2hookcallers[plugin]
        name = f'{plugin_name} v{version}'
        text = f'{name:45}  {len(hooks):3} hooks\n'
        for specname, hook_caller in self.hooks.items():
            for impl in hook_caller.get_hookimpls():
                if impl.plugin_name == plugin_name:
                    funcname = ''
                    if impl.function.__name__ != specname:
                        funcname = (
                            f'{impl.function.__module__}.'
                            f'{impl.function.__name__}'
                        )
                    text += f"  - {specname:28} {funcname}\n"
        return text


def _formatdef(func):
    return f"{func.__name__}{str(inspect.signature(func))}"


class _HookRelay:
    """Hook holder object for storing HookCaller instances.

    This object triggers (lazy) discovery of plugins as follows:  When a plugin
    hook is accessed (e.g. plugin_manager.hook.napari_get_reader), if
    ``self._needs_discovery`` is True, then it will trigger autodiscovery on
    the parent plugin_manager. Note that ``PluginManager.__init__`` sets
    ``self.hook._needs_discovery = True`` *after* hook_specifications and
    builtins have been discovered, but before external plugins are loaded.
    """

    def __init__(self, manager: PluginManager):
        self._manager = manager
        self._needs_discovery = True

    def __getattribute__(self, name) -> HookCaller:
        """Trigger manager plugin discovery when accessing hook first time."""
        if name not in ("_needs_discovery", "_manager"):
            if self._needs_discovery:
                self._manager.discover()
        return object.__getattribute__(self, name)

    def __str__(self) -> str:
        text = ''
        for hookname, hookcaller in sorted(self.items(), key=lambda x: x[0]):
            text += (
                f'{hookname:25}  {len(hookcaller.get_hookimpls()):3}'
                ' implementations\n'
            )
        return text

    def __len__(self) -> int:
        return len([k for k in vars(self) if not k.startswith("_")])

    def items(self) -> List[Tuple[str, HookCaller]]:
        """Iterate through hookcallers, removing private attributes."""
        return [
            (k, val) for k, val in vars(self).items() if not k.startswith("_")
        ]

    def values(self) -> List[HookCaller]:
        """Iterate through hookcallers, removing private attributes."""
        return [val for k, val in vars(self).items() if not k.startswith("_")]


def get_canonical_name(namespace: Any) -> str:
    """Return canonical name for a plugin object.

    Note that a plugin may be registered under a different name which was
    specified by the caller of :meth:`PluginManager.register(plugin, name)
    <.PluginManager.register>`. To obtain the name of a registered plugin
    use :meth:`get_name(plugin) <.PluginManager.get_name>` instead.
    """
    return getattr(namespace, "__name__", None) or str(id(namespace))


def iter_implementations(
    namespace, project_name: str
) -> Generator[HookImplementation, None, None]:
    # register matching hook implementations of the plugin
    for name in dir(namespace):
        # check all attributes/methods of plugin and look for functions or
        # methods that have a "{self.project_name}_impl" attribute.
        method = getattr(namespace, name)
        if not inspect.isroutine(method):
            continue
        tag = HookImplementation.format_tag(project_name)
        hookimpl_opts = getattr(method, tag, None)
        if not (isinstance(hookimpl_opts, dict) and hookimpl_opts):
            # false positive
            continue

        # create the HookImplementation instance for this method
        try:
            yield HookImplementation(method, namespace, **hookimpl_opts)
        except TypeError:
            # final fallback if the hookimpl_opts dict has invalid keys
            # it's probably not a real hook implementation anyway.
            pass


def ensure_namespace(obj: Any, name: str = 'orphan') -> Type:
    """Convert a ``dict`` to an object that provides ``getattr``.

    Parameters
    ----------
    obj : Any
        An object, may be a ``dict``, or a regular namespace object.
    name : str, optional
        A name to use for the new namespace, if created.  by default 'orphan'

    Returns
    -------
    type
        A namespace object. If ``obj`` is a ``dict``, creates a new ``type``
        named ``name``, prepopulated with the key:value pairs from ``obj``.
        Otherwise, if ``obj`` is not a ``dict``, will return the original
        ``obj``.

    Raises
    ------
    ValueError
        If ``obj`` is a ``dict`` that contains keys that are not valid
        `identifiers
        <https://docs.python.org/3.3/reference/lexical_analysis.html#identifiers>`_.
    """
    if isinstance(obj, dict):
        bad_keys = [str(k) for k in obj.keys() if not str(k).isidentifier()]
        if bad_keys:
            raise ValueError(
                f"dict contained invalid identifiers: {', '.join(bad_keys)}"
            )
        return type(name, (), obj)
    return obj


@contextmanager
def temp_path_additions(path: Optional[Union[str, List[str]]]) -> Generator:
    """A context manager that temporarily adds ``path`` to sys.path.

    Parameters
    ----------
    path : str or list of str
        A path or list of paths to add to sys.path

    Yields
    -------
    sys_path : list of str
        The current sys.path for the context.
    """
    if isinstance(path, (str, Path)):
        path = [path]
    path = [os.fspath(p) for p in path] if path else []
    to_add = [p for p in path if p not in sys.path]
    for p in to_add:
        sys.path.insert(0, p)
    try:
        yield sys.path
    finally:
        for p in to_add:
            sys.path.remove(p)


pattern = re.compile(
    r'(?P<module>[\w.]+)\s*'
    r'(:\s*(?P<attr>[\w.]+))?\s*'
    r'(?P<extras>\[.*\])?\s*$'
)


def load(value: str):
    """Load and return a module or attribute of a module (such as a class).

    If only a module is indicated by the value, return that module. Otherwise,
    return the named object.
    """
    match = pattern.match(value)
    if not match:
        raise ValueError(f"malformed entry point string: {value}")
    module = importlib.import_module(match.group('module'))
    attrs = filter(None, (match.group('attr') or '').split('.'))
    return functools.reduce(getattr, attrs, module)


def iter_available_plugins(
    group: Optional[str] = None,
    prefix: Optional[str] = None,
    path: Optional[Union[str, List[str]]] = None,
    include_uninstalled: bool = None,
) -> Generator[Tuple[str, str, Optional[str]], None, None]:
    """Discover modules by both naming convention and entry_points.

    1. `Using naming convention
        <https://packaging.python.org/guides/creating-and-discovering-plugins/#using-naming-convention>`_:
        modules installed in the environment that follow a naming convention
        (e.g. "napari_plugin"), can be discovered using :mod:`pkgutil`. This also
        enables easy discovery using the PyPI `simple API
        <https://www.python.org/dev/peps/pep-0503/>`_

    2. `Using package metadata
        <https://packaging.python.org/guides/creating-and-discovering-plugins/#using-package-metadata>`_:
        packages that declare an `entry_point
        <https://setuptools.readthedocs.io/en/latest/setuptools.html#dynamic-discovery-of-services-and-plugins>`_
        in their ``setup.py`` file that matches the ``entry_point`` argument
        can be discovered using `importlib.metadata
        <https://docs.python.org/3/library/importlib.metadata.html>`_.

    For background on entry points, see the `Entry Point specification
    <https://packaging.python.org/specifications/entry-points/>`_.

    Parameters
    ----------
    group : str
        The entry_point group name to search for
    prefix : str
        Any modules found in sys.path whose names begin with ``prefix``
        will be imported and searched for hook implementations.
    path : str or list of str, optional
        Path or paths to add to sys.path before importing, removed at the end.
    include_uninstalled : bool, optional
        Whether to search for "local" (uninstalled) modules.  Requires that a
        prefix is provided.  By default, True when prefix is provided.

    Raises
    ------
    ValueError
        If ``include_uninstalled`` is true and ``prefix`` is not provided.
    """
    if include_uninstalled is None:
        include_uninstalled = bool(prefix)
    with temp_path_additions(path):
        _seen = set()
        for dist in importlib_metadata.distributions():
            matched = False
            if group and not os.getenv("DISABLE_ENTRYPOINT_PLUGINS"):
                for ep in dist.entry_points:
                    if ep.group == group:  # type: ignore
                        matched = True
                        _seen.add(ep.value.split(".", maxsplit=1)[0])
                        yield (
                            ep.name,
                            ep.value,
                            dist.metadata.get("name"),
                        )
            if matched:
                continue
            if prefix and not os.getenv("DISABLE_PREFIX_PLUGINS"):
                name = dist.metadata.get("name")
                if not name or name == prefix or (not name.startswith(prefix)):
                    continue
                top_modules = dist.read_text('top_level.txt') or ""
                for mod in filter(None, top_modules.split('\n')):
                    if mod.startswith(prefix):
                        _seen.add(mod)
                        yield (name, mod, name)

        if include_uninstalled and not os.getenv("DISABLE_PREFIX_PLUGINS"):
            from pkgutil import iter_modules

            if not prefix:
                raise ValueError(
                    "A prefix must be provided with 'include_uninstalled'."
                )
            for finder, mod_name, ispkg in iter_modules():
                if (
                    mod_name.startswith(prefix)
                    and mod_name != prefix
                    and mod_name not in _seen
                ):
                    yield (mod_name, mod_name, None)



try:
    from ._version import version as __version__
except ImportError:
    # broken installation, we don't even try
    # unknown only works because we do poor mans version compare
    __version__ = "unknown"

__all__ = [
    "get_metadata",
    "HookCaller",
    "HookCallError",
    "HookImplementation",
    "HookImplementationMarker",
    "HookResult",
    "HookSpecification",
    "HookSpecificationMarker",
    "napari_hook_implementation",
    "napari_hook_specification",
    "PluginCallError",
    "PluginError",
    "PluginImplementationError",
    "PluginImportError",
    "PluginManager",
    "PluginRegistrationError",
    "PluginValidationError",
    "standard_metadata",
]

from .callers import HookResult
from .dist import get_metadata, standard_metadata
from .exceptions import (
    HookCallError,
    PluginCallError,
    PluginError,
    PluginImplementationError,
    PluginImportError,
    PluginRegistrationError,
    PluginValidationError,
)
from .hooks import HookCaller
from .implementation import HookImplementation, HookSpecification
from .manager import PluginManager
from .markers import HookImplementationMarker, HookSpecificationMarker

napari_hook_implementation = HookImplementationMarker("napari")
napari_hook_specification = HookSpecificationMarker("napari")



import os

import pytest

from napari_plugin_engine import (
    PluginError,
    PluginImportError,
    PluginValidationError,
)
from napari_plugin_engine.dist import get_version, standard_metadata
from napari_plugin_engine.manager import temp_path_additions

GOOD_PLUGIN = """
from napari_plugin_engine import HookImplementationMarker

@HookImplementationMarker("test")
def test_specification(arg1, arg2):
    return arg1 + arg2
"""

INVALID_PLUGIN = """
from napari_plugin_engine import HookImplementationMarker

@HookImplementationMarker("test")
def test_specification(arg1, arg2, arg3):
    return arg1 + arg2 + arg3
"""


@pytest.fixture
def app_good_plugin(tmp_path):
    """A good plugin with a name prefix we will search for."""
    (tmp_path / "app_good_plugin.py").write_text(GOOD_PLUGIN)


@pytest.fixture
def good_entrypoint_plugin(tmp_path):
    """A good plugin that uses entry points."""
    (tmp_path / "good_entrypoint_plugin.py").write_text(GOOD_PLUGIN)
    distinfo = tmp_path / "good_entrypoint_plugin-1.2.3.dist-info"
    distinfo.mkdir()
    (distinfo / "top_level.txt").write_text('good_entrypoint_plugin')
    (distinfo / "entry_points.txt").write_text(
        "[app.plugin]\ngood_entry = good_entrypoint_plugin"
    )
    (distinfo / "METADATA").write_text(
        "Metadata-Version: 2.1\n"
        "Name: good_entry\n"
        "Version: 1.2.3\n"
        "Author-Email: example@example.com\n"
        "Home-Page: https://www.example.com\n"
        "Requires-Python: >=3.6\n"
    )


@pytest.fixture
def double_convention_plugin(tmp_path):
    """A good plugin that uses entry points but ALSO has naming convention"""
    module_folder = tmp_path / "app_double_plugin"
    module_folder.mkdir()
    (module_folder / "__init__.py").write_text('')
    (module_folder / "module_a.py").write_text(GOOD_PLUGIN)
    (module_folder / "module_b.py").write_text(GOOD_PLUGIN)
    distinfo = tmp_path / "app_double_plugin-3.2.1.dist-info"
    distinfo.mkdir()
    (distinfo / "top_level.txt").write_text('app_double_plugin')
    (distinfo / "entry_points.txt").write_text(
        "[app.plugin]\n"
        "double_a = app_double_plugin.module_a\n"
        "double_b = app_double_plugin.module_b\n"
    )
    (distinfo / "METADATA").write_text(
        "Metadata-Version: 2.1\n"
        "Name: double-package\n"
        "Version: 3.2.1\n"
        "Author-Email: example@example.com\n"
        "Home-Page: https://www.example.com\n"
        "Requires-Python: >=3.6\n"
    )


@pytest.fixture
def invalid_entrypoint_plugin(tmp_path):
    """A good plugin that uses entry points."""
    (tmp_path / "invalid_entrypoint_plugin.py").write_text(INVALID_PLUGIN)
    distinfo = tmp_path / "invalid_entrypoint_plugin-1.2.3.dist-info"
    distinfo.mkdir()
    (distinfo / "top_level.txt").write_text('invalid_entrypoint_plugin')
    (distinfo / "entry_points.txt").write_text(
        "[app.plugin]\ninvalid = invalid_entrypoint_plugin"
    )
    (distinfo / "METADATA").write_text(
        "Metadata-Version: 2.1\n"
        "Name: invalid\n"
        "Version: 1.2.3\n"
        "Author-Email: example@example.com\n"
        "Home-Page: https://www.example.com\n"
        "Requires-Python: >=3.6\n"
    )
    return tmp_path


@pytest.fixture
def app_invalid_plugin(tmp_path):
    (tmp_path / "app_invalid_plugin.py").write_text(INVALID_PLUGIN)


@pytest.fixture
def app_broken_plugin(tmp_path):
    (tmp_path / "app_broken_plugin.py").write_text('raise ValueError("broke")')


@pytest.fixture
def full_plugin_manager(
    tmp_path,
    add_specification,
    test_plugin_manager,
    good_entrypoint_plugin,
    invalid_entrypoint_plugin,
    app_good_plugin,
    app_invalid_plugin,
    double_convention_plugin,
):
    @add_specification
    def test_specification(arg1, arg2): ...

    test_plugin_manager.discover_entry_point = 'app.plugin'
    test_plugin_manager.discover_prefix = 'app_'
    test_plugin_manager.discover_path = tmp_path

    cnt, err = test_plugin_manager.discover()
    assert cnt == 4
    assert len(err) == 2
    return test_plugin_manager


def test_plugin_meta(
    tmp_path,
    add_specification,
    test_plugin_manager,
    app_good_plugin,
    good_entrypoint_plugin,
    double_convention_plugin,
):

    test_plugin_manager.discover_entry_point = 'app.plugin'
    test_plugin_manager.discover_prefix = 'app_'

    with temp_path_additions(tmp_path):

        cnt, err = test_plugin_manager.discover()
        assert set(test_plugin_manager.plugins) == {
            'double_a',
            'double_b',
            'good_entry',
            'app_good_plugin',
        }

        versions = {
            'double_a': '3.2.1',
            'double_b': '3.2.1',
            'good_entry': '1.2.3',
            'app_good_plugin': '',
        }
        for name, plug in test_plugin_manager.plugins.items():
            assert versions[name] == get_version(plug)
            if name == 'app_good_plugin':
                # this one doesn't have any metadata.. but it will have plugin_name
                assert standard_metadata(plug) == {}
            else:
                assert get_version(plug) == standard_metadata(plug).get(
                    'version'
                )


@pytest.mark.parametrize(
    'regkwargs',
    [
        {},
        {'entry_point': 'app.plugin'},
        {'prefix': 'app_'},
        {'entry_point': 'app.plugin', 'prefix': 'app_'},
    ],
    ids=['neither', 'entry_point', 'prefix', 'both'],
)
def test_double_convention(
    regkwargs,
    tmp_path,
    add_specification,
    test_plugin_manager,
    double_convention_plugin,
):
    """Plugins using both naming convention and entrypoints only register once."""

    @add_specification
    def test_specification(arg1, arg2): ...

    assert not test_plugin_manager.plugins

    if 'entry_point' in regkwargs:
        test_plugin_manager.discover_entry_point = regkwargs['entry_point']
    if 'prefix' in regkwargs:
        test_plugin_manager.discover_prefix = regkwargs['prefix']
    test_plugin_manager.discover_path = tmp_path

    cnt, _ = test_plugin_manager.discover()
    hook_caller = test_plugin_manager.hook.test_specification
    plugin_names = list(test_plugin_manager.plugins.keys())
    if regkwargs:
        if 'entry_point' in regkwargs:
            # if an entry_point with a matching group is provided
            # the plugin will be named after the entrypoint name
            assert 'double_a' in plugin_names
            assert 'double_b' in plugin_names
            assert 'double-package' not in test_plugin_manager.plugins
            assert len(hook_caller.get_hookimpls()) == 2
        else:  # just prefix
            # if entry point discovery is disabled, but a top-level module
            # matches the naming convnetion, then the plugin will be named using
            # the top level module
            assert 'app_double_plugin' in plugin_names
            # however, in this case, there are no implementations in the
            # top level module (which is the only way naming convention
            # works)
            assert len(hook_caller.get_hookimpls()) == 0
    else:
        assert cnt == 0
        assert not hook_caller.get_hookimpls()
        assert 'double' not in plugin_names


def test_plugin_discovery_by_prefix(
    tmp_path,
    add_specification,
    test_plugin_manager,
    app_good_plugin,
    app_invalid_plugin,
):
    """Make sure discovery by package prefix works"""

    @add_specification
    def test_specification(arg1, arg2): ...

    assert test_plugin_manager.hook.test_specification.spec
    assert not test_plugin_manager.plugins

    test_plugin_manager.discover_prefix = 'app_'
    test_plugin_manager.discover_path = tmp_path

    # discover modules that begin with `app_`
    count, errs = test_plugin_manager.discover()
    # we should have had one success and one error.
    assert count == len(errs) == 1

    # the app_good_plugin module should have been found, with one hookimpl
    assert 'app_good_plugin' in test_plugin_manager.plugins.keys()
    impls = test_plugin_manager.hook.test_specification.get_hookimpls()
    assert 'app_good_plugin' in [i.plugin_name for i in impls]

    # the plugin with the invalid spec is in the path that we loaded
    assert 'app_invalid_plugin.py' in os.listdir(tmp_path)
    # but it wasn't added to the plugin manager
    assert 'app_invalid_plugin' not in test_plugin_manager.plugins.keys()
    # However an error should have been logged for the invalid plugin
    assert not test_plugin_manager.get_errors('app_good_plugin')
    errs = test_plugin_manager.get_errors('app_invalid_plugin')
    assert errs
    assert isinstance(errs[0], PluginValidationError)
    # and it should now be blocked
    # assert test_plugin_manager.is_blocked('app_invalid_plugin')

    # if we unblock the plugin and turn off ignore_errors
    # we'll get a registration error at discovery
    test_plugin_manager.set_blocked('app_invalid_plugin', False)
    with pytest.raises(PluginValidationError):
        test_plugin_manager.discover(ignore_errors=False)


def test_plugin_discovery_by_prefix_with_bad_plugin(
    tmp_path, add_specification, test_plugin_manager, app_broken_plugin
):
    """Make sure bad plugins can raise errors"""

    test_plugin_manager.discover_prefix = 'app_'
    test_plugin_manager.discover_path = tmp_path

    with pytest.raises(PluginImportError):
        test_plugin_manager.discover(ignore_errors=False)


def test_plugin_discovery_by_entry_point(
    tmp_path,
    add_specification,
    test_plugin_manager,
    good_entrypoint_plugin,
    invalid_entrypoint_plugin,
):
    @add_specification
    def test_specification(arg1, arg2): ...

    hook_caller = test_plugin_manager.hook.test_specification
    assert hook_caller.spec
    assert not test_plugin_manager.plugins

    test_plugin_manager.discover_entry_point = 'app.plugin'
    test_plugin_manager.discover_path = tmp_path

    # discover modules that begin with `app_`
    cnt, err = test_plugin_manager.discover()
    # we should have had one success and one error.
    assert cnt == len(err) == 1

    # the app_good_plugin module should have been found, with one hookimpl
    assert 'good_entry' in test_plugin_manager.plugins.keys()
    assert 'good_entry' in [i.plugin_name for i in hook_caller.get_hookimpls()]

    # the plugin with the invalid spec is in the path that we loaded
    assert 'invalid_entrypoint_plugin.py' in os.listdir(tmp_path)
    # but it wasn't added to the plugin manager
    assert 'invalid' not in test_plugin_manager.plugins.keys()
    # However an error should have been logged for the invalid plugin
    assert not test_plugin_manager.get_errors('good_entry')
    errs = test_plugin_manager.get_errors('invalid')
    assert errs
    assert isinstance(errs[0], PluginValidationError)
    # and it should now be blocked
    # assert test_plugin_manager.is_blocked('invalid')

    # if we unblock the plugin and turn off ignore_errors
    # we'll get a registration error at discovery
    test_plugin_manager.set_blocked('invalid', False)
    with pytest.raises(PluginValidationError):
        test_plugin_manager.discover(ignore_errors=False)


def test_identical_entry_point_discovery(
    good_entrypoint_plugin, test_plugin_manager, tmp_path
):
    test_plugin_manager.discover_entry_point = 'app.plugin'
    test_plugin_manager.discover_path = tmp_path

    # make second plugin with same entry point
    (tmp_path / "good_entrypoint_plugin2.py").write_text(GOOD_PLUGIN)
    distinfo = tmp_path / "good_entrypoint_plugin2-1.2.3.dist-info"
    distinfo.mkdir()
    (distinfo / "top_level.txt").write_text('good_entrypoint_plugin2')
    (distinfo / "entry_points.txt").write_text(
        "[app.plugin]\ngood_entry = good_entrypoint_plugin2"
    )
    (distinfo / "METADATA").write_text(
        "Metadata-Version: 2.1\n"
        "Name: good_entry\n"
        "Version: 1.2.3\n"
        "Author-Email: example@example.com\n"
        "Home-Page: https://www.example.com\n"
        "Requires-Python: >=3.6\n"
    )

    with pytest.warns(UserWarning):
        cnt, err = test_plugin_manager.discover()
        assert cnt == 2
        assert 'good_entry-2' in test_plugin_manager.plugins
        assert not err

    # test discovering again doesn't rediscover same plugins
    cnt, err = test_plugin_manager.discover()
    assert cnt == 0
    assert not err
    assert 'good_entry' in test_plugin_manager.plugins
    assert 'good_entry-2' in test_plugin_manager.plugins
    assert len(test_plugin_manager.plugins) == 2


def test_lazy_autodiscovery(
    tmp_path, add_specification, test_plugin_manager, good_entrypoint_plugin
):
    test_plugin_manager.discover_entry_point = 'app.plugin'
    assert test_plugin_manager.hook._needs_discovery is True
    with test_plugin_manager.discovery_blocked():

        @add_specification
        def test_specification(arg1, arg2): ...

    assert test_plugin_manager.hook._needs_discovery is True

    assert not test_plugin_manager.plugins
    with temp_path_additions(tmp_path):
        hook_caller = test_plugin_manager.hook.test_specification

    assert hook_caller.spec
    assert test_plugin_manager.plugins.get('good_entry')
    assert test_plugin_manager.hook._needs_discovery is False


def test_discovery_all_together(full_plugin_manager):
    hook_caller = full_plugin_manager.hook.test_specification
    assert len(hook_caller.get_hookimpls()) == 4
    assert len(full_plugin_manager.plugins) == 4
    assert 'double_a' in full_plugin_manager.plugins.keys()
    assert 'double_b' in full_plugin_manager.plugins.keys()
    assert 'app_good_plugin' in full_plugin_manager.plugins.keys()
    assert 'good_entry' in full_plugin_manager.plugins.keys()


def test_getting_errors(invalid_entrypoint_plugin, caplog):
    with temp_path_additions(invalid_entrypoint_plugin):
        import invalid_entrypoint_plugin as mod

    try:
        raise ValueError('I caused this')
    except ValueError as e:
        err = PluginError(plugin=mod, plugin_name='invalid', cause=e)
    errs = PluginError.get(plugin=mod)
    assert mod in {p.plugin for p in errs}
    errs = PluginError.get(plugin_name='invalid')
    assert 'invalid' in {p.plugin_name for p in errs}

    assert 'I caused this' in err.format()
    err.log()
    assert ' Error in plugin "invalid"' in caplog.text


@pytest.mark.parametrize('blocked', ['ALL', 'ENTRYPOINT', 'PREFIX'])
def test_env_var_disable(
    blocked,
    tmp_path,
    add_specification,
    test_plugin_manager,
    good_entrypoint_plugin,
    app_good_plugin,
    monkeypatch,
):
    @add_specification
    def test_specification(arg1, arg2): ...

    test_plugin_manager.discover_entry_point = 'app.plugin'
    test_plugin_manager.discover_prefix = 'app_'
    test_plugin_manager.discover_path = tmp_path

    monkeypatch.setenv(f'DISABLE_{blocked}_PLUGINS', '1')
    if blocked == 'ALL':
        with pytest.warns(UserWarning):
            cnt, err = test_plugin_manager.discover()
    else:
        cnt, err = test_plugin_manager.discover()
    assert cnt == (0 if blocked == 'ALL' else 1)

    if blocked == 'ENTRYPOINT':
        assert 'app_good_plugin' in test_plugin_manager.plugins.keys()
        assert 'good_entry' not in test_plugin_manager.plugins.keys()
    elif blocked == 'PREFIX':
        assert 'app_good_plugin' not in test_plugin_manager.plugins.keys()
        assert 'good_entry' in test_plugin_manager.plugins.keys()
    elif blocked == 'ALL':
        assert 'app_good_plugin' not in test_plugin_manager.plugins.keys()
        assert 'good_entry' not in test_plugin_manager.plugins.keys()

    monkeypatch.delenv(f'DISABLE_{blocked}_PLUGINS')



import logging
from enum import Enum
from types import TracebackType
from typing import TYPE_CHECKING, Any, List, Optional, Tuple, Type, Union

from .dist import standard_metadata

if TYPE_CHECKING:
    from .manager import PluginManager  # noqa: F401


ExcInfoTuple = Tuple[Type[Exception], Exception, Optional[TracebackType]]


# https://www.python.org/dev/peps/pep-0484/#support-for-singleton-types-in-unions
class Empty(Enum):
    token = 0


_empty = Empty.token


class PluginError(Exception):
    """Base class for exceptions relating to plugins.

    Parameters
    ----------
    message : str, optional
        An optional error message, by default ''
    namespace : Optional[Any], optional
        The python object that caused the error, by default None
    cause : Exception, optional
        Exception that caused the error. Same as ``raise * from``.
        by default None
    """

    _record: List['PluginError'] = []

    def __init__(
        self,
        message: str = '',
        *,
        plugin: Optional[Any] = None,
        plugin_name: Optional[str] = None,
        cause: Optional[BaseException] = None,
    ):
        self.plugin = plugin
        self.plugin_name = plugin_name
        if not message:
            name = plugin_name or getattr(plugin, '__name__', str(id(plugin)))
            message = f'Error in plugin "{name}"'
            if cause:
                message += f': {cause}'
        super().__init__(message)
        self.__cause__ = cause
        # store all PluginError instances.  can be retrieved with get()
        PluginError._record.append(self)

    @classmethod
    def get(
        cls,
        *,
        plugin: Union[Any, Empty] = _empty,
        plugin_name: Union[str, Empty] = _empty,
        error_type: Union[Type['PluginError'], Empty] = _empty,
    ) -> List['PluginError']:
        """Return errors that have been logged, filtered by parameters.

        Parameters
        ----------
        manager : PluginManager, optional
            If provided, will restrict errors to those that are owned by
            ``manager``.
        plugin_name : str
            If provided, will restrict errors to those that were raised by
            ``plugin_name``.
        error_type : Exception
            If provided, will restrict errors to instances of ``error_type``.

        Returns
        -------
        list of PluginError
            A list of PluginErrors that have been instantiated during this
            session that match the provided parameters.

        Raises
        ------
        TypeError
            If ``error_type`` is provided and is not an exception class.
        """
        errors: List['PluginError'] = []
        for error in cls._record:
            if plugin is not _empty and error.plugin != plugin:
                continue
            if plugin_name is not _empty and error.plugin_name != plugin_name:
                continue
            if error_type is not _empty:
                import inspect

                if not (
                    inspect.isclass(error_type)
                    and issubclass(error_type, BaseException)
                ):
                    raise TypeError(
                        "The `error_type` argument must be an exception class"
                    )
                if not isinstance(error.__cause__, error_type):
                    continue
            errors.append(error)
        return errors

    def format(self, package_info: bool = True):
        msg = f'PluginError: {self}'
        if self.__cause__:
            msg = msg.replace(str(self.__cause__), '').strip(": ") + "\n"
            cause = repr(self.__cause__).replace("\n", "\n" + " " * 13)
            msg += f'  Cause was: {cause}'

            # show the exact file and line where the error occured
            cause_tb = self.__cause__.__traceback__
            if cause_tb:
                while True:
                    if not cause_tb.tb_next:
                        break
                    cause_tb = cause_tb.tb_next

                msg += f'\n    in file: {cause_tb.tb_frame.f_code.co_filename}'
                msg += f'\n    at line: {cause_tb.tb_lineno}'
        else:
            msg += "\n"

        if package_info and self.plugin:
            try:
                meta = standard_metadata(self.plugin)
                meta.pop('license', None)
                meta.pop('summary', None)
                if meta:
                    msg += "\n" + "\n".join(
                        [
                            f'{k: >11}: {v}'
                            for k, v in sorted(meta.items())
                            if v
                        ]
                    )
            except ValueError:
                pass
        msg += '\n'
        return msg

    def log(
        self,
        package_info: bool = True,
        logger: Union[logging.Logger, None, str] = None,
        level: int = logging.ERROR,
    ):
        """Log this error with metadata, optionally provide logger and level.

        Parameters
        ----------
        package_info : bool, optional
            If true, will include package metadata in log, by default True
        logger : logging.Logger or str, optional
            A Logger instance or name of a logger to use, by default None
        level : int, optional
            The logging level to use, by default logging.ERROR
        """
        if not isinstance(logger, logging.Logger):
            logger = logging.getLogger(logger)

        logger.log(level, self.format(package_info=package_info))

    def info(self) -> ExcInfoTuple:
        """Return info as would be returned from sys.exc_info()."""
        return (self.__class__, self, self.__traceback__)


class HookCallError(PluginError):
    """If a hook is called incorrectly.

    Usually this results when a HookCaller is called without the appropriate
    arguments.
    """


class PluginImportError(PluginError, ImportError):
    """Plugin module is unimportable."""


class PluginRegistrationError(PluginError):
    """If an unexpected error occurs during registration."""


class PluginImplementationError(PluginError):
    """Base class for errors pertaining to a specific hook implementation."""

    def __init__(self, hook_implementation, msg=None, cause=None):
        plugin = hook_implementation.plugin
        plugin_name = hook_implementation.plugin_name
        specname = hook_implementation.specname

        if not msg:
            msg = f"Error in plugin '{plugin_name}', hook '{specname}'"
            if cause:
                msg += f": {str(cause)}"

        super().__init__(
            msg,
            plugin=plugin,
            plugin_name=plugin_name,
            cause=cause,
        )


class PluginValidationError(PluginImplementationError):
    """When a plugin implementation fails validation."""


class PluginCallError(PluginImplementationError):
    """Raised when an error is raised when calling a plugin implementation."""



"""Call loop machinery."""

import sys
from types import TracebackType
from typing import Any, List, Optional, Tuple, Type, Union

from .exceptions import HookCallError, PluginCallError
from .implementation import HookImplementation


def _raise_wrapfail(wrap_controller, msg):
    co = wrap_controller.gi_code
    raise RuntimeError(
        "wrap_controller at %r %s:%d %s"
        % (co.co_name, co.co_filename, co.co_firstlineno, msg)
    )


ExcInfo = Union[
    Tuple[Type[BaseException], BaseException, TracebackType],
    Tuple[None, None, None],
]


class HookResult:
    """A class to store/modify results from a :func:`~hooks._multicall` hook loop.

    Results are accessed in ``.result`` property, which will also raise
    any exceptions that occured during the hook loop.

    Parameters
    ----------
    results : List[Tuple[Any, HookImplementation]]
        A list of (result, HookImplementation) tuples, with the result and
        HookImplementation object responsible for each result collected during
        a _multicall loop.
    excinfo : tuple
        The output of sys.exc_info() if raised during the multicall loop.
    firstresult : bool, optional
        Whether the hookspec had ``firstresult == True``, by default False.
        If True, self._result, and self.implementation will be single values,
        otherwise they will be lists.
    plugin_errors : list
        A list of any :class:`PluginCallError` instances that were created
        during the multicall loop.
    """

    def __init__(
        self,
        result: List[Tuple[Any, HookImplementation]],
        excinfo: Optional[ExcInfo],
        firstresult: bool = False,
        plugin_errors: Optional[List[PluginCallError]] = None,
    ):
        self._result: Any = []
        #: The HookImplementation(s) that were responsible for each result in ``result``
        self.implementation: Optional[
            Union[HookImplementation, List[HookImplementation]]
        ] = []
        #: Whether this HookResult came from a ``firstresult`` multicall.
        self.is_firstresult: bool = firstresult
        self._excinfo = excinfo
        self.plugin_errors = plugin_errors

        if result:
            self._result, self.implementation = tuple(zip(*result))
            self._result = list(self._result)
        if firstresult:
            if self._result:
                self._result = self._result[0]
                self.implementation = self.implementation[0]  # type: ignore
            else:
                self._result = None
                self.implementation = None

        #: Name of last hookwrapper that changed the result, if any
        self._modified_by: Optional[str] = None

    @property
    def excinfo(self):
        return self._excinfo

    @classmethod
    def from_call(cls, func):
        """Used when hookcall monitoring is enabled.

        https://pluggy.readthedocs.io/en/latest/#call-monitoring
        """
        __tracebackhide__ = True
        try:
            return func()
        except BaseException:
            return cls(None, sys.exc_info())

    def force_result(self, result: Any):
        """Force the result(s) to ``result``.

        This may be used by hookwrappers to alter this result object.

        If the hook was marked as a ``firstresult`` a single value should
        be set otherwise set a (modified) list of results. Any exceptions
        found during invocation will be deleted.
        """
        import inspect

        self._result = result
        self._excinfo = None
        self._modified_by = inspect.stack()[1].function

    @property
    def result(self) -> Union[Any, List[Any]]:
        """Return the result(s) for this hook call.

        If the hook was marked as a ``firstresult`` only a single value
        will be returned otherwise a list of results.
        """
        __tracebackhide__ = True
        if self._excinfo is not None:
            _type, value, traceback = self._excinfo
            if value:
                raise value.with_traceback(traceback)
        return self._result


def _multicall(
    hook_impls: List[HookImplementation],
    caller_kwargs: dict,
    firstresult: bool = False,
) -> HookResult:
    """The primary :class:`~napari_plugin_engine.HookImplementation` call loop.

    Parameters
    ----------
    hook_impls : list
        A sequence of hook implementation (HookImplementation) objects
    caller_kwargs : dict
        Keyword:value pairs to pass to each ``hook_impl.function``.  Every
        key in the dict must be present in the ``argnames`` property for each
        ``hook_impl`` in ``hook_impls``.
    firstresult : bool, optional
        If ``True``, return the first non-null result found, otherwise, return
        a list of results from all hook implementations, by default False

    Returns
    -------
    outcome : HookResult
        A :class:`HookResult` object that contains the results returned by
        plugins along with other metadata about the call.

    Raises
    ------
    HookCallError
        If one or more of the keys in ``caller_kwargs`` is not present in one
        of the ``hook_impl.argnames``.
    PluginCallError
        If ``firstresult == True`` and a plugin raises an Exception.
    """
    __tracebackhide__ = True
    results = []
    errors: List['PluginCallError'] = []
    excinfo: Optional[ExcInfo] = None
    try:  # run impl and wrapper setup functions in a loop
        teardowns = []
        try:
            for hook_impl in reversed(hook_impls):
                # skip disabled hook implementations
                if not getattr(hook_impl, 'enabled', True):
                    continue
                args: List[Any] = []
                try:
                    args = [
                        caller_kwargs[argname]
                        for argname in hook_impl.argnames
                    ]
                except KeyError:
                    raise HookCallError(
                        "hook call must provide argument the following "
                        f"arguments: {set(hook_impl.argnames)!r}, but provided"
                        f" {set(caller_kwargs)!r}"
                    )

                if hook_impl.hookwrapper:
                    try:
                        gen = hook_impl(*args)
                        next(gen)  # first yield
                        teardowns.append(gen)
                    except StopIteration:
                        _raise_wrapfail(gen, "did not yield")
                else:
                    res = None
                    # this is where the plugin function actually gets called
                    # we put it in a try/except so that if one plugin throws
                    # an exception, we don't lose the whole loop
                    try:
                        res = hook_impl(*args)
                    except Exception as exc:
                        # creating a PluginCallError will store it for later
                        # in plugins.exceptions.PLUGIN_ERRORS
                        errors.append(PluginCallError(hook_impl, cause=exc))
                        # if it was a `firstresult` hook, break and raise now.
                        if firstresult:
                            break

                    if res is not None:
                        results.append((res, hook_impl))
                        if firstresult:  # halt further impl calls
                            break
        except BaseException:
            excinfo = sys.exc_info()
    finally:
        if firstresult and errors:
            raise errors[-1]

    outcome = HookResult(
        results,
        excinfo=excinfo,
        firstresult=firstresult,
        plugin_errors=errors,
    )

    # run all wrapper post-yield blocks
    for gen in reversed(teardowns):
        try:
            gen.send(outcome)
            _raise_wrapfail(gen, "has second yield")
        except StopIteration:
            pass

    return outcome



import pytest

from napari_plugin_engine import (
    HookImplementationMarker,
    HookSpecificationMarker,
    PluginManager,
)

dummy_hook_implementation = HookImplementationMarker("dummy")
dummy_hook_specification = HookSpecificationMarker("dummy")


class MySpec:
    @dummy_hook_specification
    def myhook(self):
        pass


class Plugin_1:
    @dummy_hook_implementation
    def myhook(self):
        return "p1"


class Plugin_2:
    @dummy_hook_implementation(tryfirst=True)
    def myhook(self):
        return "p2"


class Plugin_3:
    @dummy_hook_implementation
    def myhook(self):
        return "p3"


class Wrapper:
    @dummy_hook_implementation(hookwrapper=True)
    def myhook(self):
        yield


p1, p2, p3, wrapper = Plugin_1(), Plugin_2(), Plugin_3(), Wrapper()


@pytest.fixture
def dummy_plugin_manager():
    plugin_manager = PluginManager("dummy")
    plugin_manager.add_hookspecs(MySpec)
    plugin_manager.register(p1, name='p1')
    plugin_manager.register(p2, name='p2')
    plugin_manager.register(p3, name='p3')
    plugin_manager.register(wrapper, name='wrapper')
    return plugin_manager


# p2 is first because it was declared with tryfirst=True
# p3 is second because of "last-in-first-out" order
START_ORDER = ['p2', 'p3', 'p1']


@pytest.mark.parametrize(
    'order, expected_result',
    [
        ([], START_ORDER),
        (['p2'], START_ORDER),
        (['p2', 'p3'], START_ORDER),
        (['p1', 'p2', 'p3'], ['p1', 'p2', 'p3']),
        (['p1', 'p3', 'p2'], ['p1', 'p3', 'p2']),
        (['p1', 'p3'], ['p1', 'p3', 'p2']),
        (['p1'], ['p1', 'p2', 'p3']),
        (['p3'], ['p3', 'p2', 'p1']),
    ],
)
def test_reordering_hook_caller(dummy_plugin_manager, order, expected_result):
    """Test that the permute_hook_implementations function reorders hooks."""
    hook_caller = dummy_plugin_manager.hooks.myhook

    assert hook_caller() == START_ORDER
    hook_caller.bring_to_front(order)
    assert hook_caller() == expected_result
    # return to original order
    hook_caller.bring_to_front(START_ORDER)
    assert hook_caller() == START_ORDER

    # try again using HookImplementation INSTANCES instead of plugin names
    instances = [hook_caller.get_plugin_implementation(i) for i in order]
    hook_caller.bring_to_front(instances)
    assert hook_caller() == expected_result


def test_reordering_hook_caller_raises(dummy_plugin_manager):
    """Test that invalid calls to permute_hook_implementations raise errors."""
    hook_caller = dummy_plugin_manager.hooks.myhook

    with pytest.raises(TypeError):
        # all items must be the name of a plugin, or a HookImplementation instance
        hook_caller.bring_to_front([1, 2])

    with pytest.raises(ValueError):
        # 'wrapper' is the name of a plugin that provides an implementation...
        # but it is a hookwrappers which is not valid for `bring_to_front`
        hook_caller.bring_to_front(['p1', 'wrapper'])

    with pytest.raises(ValueError):
        # 'p4' is not in the list
        hook_caller.bring_to_front(['p1', 'p4'])

    with pytest.raises(ValueError):
        # duplicate entries are not allowed
        hook_caller.bring_to_front(['p1', 'p1', 'p2'])

    with pytest.raises(ValueError):
        # too many values
        hook_caller.bring_to_front(['p1', 'p1', 'p2', 'p4', 'p3', 'p1'])

    with pytest.raises(TypeError):
        # it has to be a list
        hook_caller.bring_to_front('p1')


def test_hook_caller_kwargs(dummy_plugin_manager):
    hook_caller = dummy_plugin_manager.hooks.myhook
    assert hook_caller() == ['p2', 'p3', 'p1']

    for p in ['p2', 'p3', 'p1']:
        # call with a specific plugin
        assert hook_caller(_plugin=p) == p
        impl = hook_caller.get_plugin_implementation(p)
        # call without specific plugins/impls
        expected = ['p2', 'p3', 'p1']
        expected.remove(p)
        assert hook_caller(_skip_impls=[impl]) == expected


def test_disable_impls(dummy_plugin_manager):
    hook_caller = dummy_plugin_manager.hooks.myhook
    assert hook_caller() == ['p2', 'p3', 'p1']

    for p in ['p2', 'p3', 'p1']:
        # call with a specific plugin
        expected = ['p2', 'p3', 'p1']
        assert hook_caller() == expected
        assert hook_caller(_plugin=p) == p
        assert hook_caller(_plugin=p) == hook_caller._call_plugin(p)
        impl = hook_caller.get_plugin_implementation(p)
        # call without specific plugins/impls
        expected.remove(p)
        assert hook_caller(_skip_impls=[impl]) == expected
        hook_caller.disable_plugin(p)
        assert hook_caller() == expected
        hook_caller.enable_plugin(p)
        assert set(hook_caller()) == set([p] + expected)

    with pytest.raises(TypeError):
        # cannot call wrappers directly
        hook_caller(_plugin='wrapper')


def test_get_plugin_implementation(dummy_plugin_manager):
    hook_caller = dummy_plugin_manager.hooks.myhook
    impl = hook_caller.get_plugin_implementation('p1')
    assert impl
    assert impl.plugin_name == 'p1'

    with pytest.raises(KeyError):
        hook_caller.get_plugin_implementation('pasdfsdf1')



import warnings

import pytest

from napari_plugin_engine import (
    HookImplementationMarker,
    HookSpecificationMarker,
    PluginManager,
)

hookspec = HookSpecificationMarker("example")
hookimpl = HookImplementationMarker("example")


def test_warn_when_deprecated_specified(recwarn):
    warning = DeprecationWarning("foo is deprecated")

    class Spec:
        @hookspec(warn_on_impl=warning)
        def foo(self):
            pass

    class Plugin:
        @hookimpl
        def foo(self):
            pass

    pm = PluginManager(hookspec.project_name)
    pm.add_hookspecs(Spec)

    with pytest.warns(DeprecationWarning) as records:
        pm.register(Plugin())
    (record,) = records
    assert record.message is warning
    assert record.filename == Plugin.foo.__code__.co_filename
    assert record.lineno == Plugin.foo.__code__.co_firstlineno


def test_plugin_getattr_raises_errors():
    """napari_plugin_engine must be able to handle plugins which raise weird exceptions
    when getattr() gets called (#11).
    """

    class DontTouchMe:
        def __getattr__(self, x):
            raise Exception("cant touch me")

    class Module:
        pass

    module = Module()
    module.x = DontTouchMe()

    pm = PluginManager(hookspec.project_name)
    # register() would raise an error
    pm.register(module, "donttouch")
    assert pm.plugins.get("donttouch") is module


def test_warning_on_call_vs_hookspec_arg_mismatch():
    """Verify that is a hook is called with less arguments then defined in the
    spec that a warning is emitted.
    """

    class Spec:
        @hookspec
        def myhook(self, arg1, arg2):
            pass

    class Plugin:
        @hookimpl
        def myhook(self, arg1):
            pass

    pm = PluginManager(hookspec.project_name)
    pm.register(Plugin())
    pm.add_hookspecs(Spec())

    with warnings.catch_warnings(record=True) as warns:
        warnings.simplefilter("always")

        # calling should trigger a warning
        pm.hook.myhook(arg1=1)

        assert len(warns) == 1
        warning = warns[-1]
        assert issubclass(warning.category, Warning)
        assert "Argument(s) ('arg2',)" in str(warning.message)


def test_repr():
    class Hook:
        @hookspec
        def myhook(arg): ...

    class Plugin:
        @hookimpl
        def myhook(arg):
            raise NotImplementedError()

    pm = PluginManager(hookspec.project_name)
    pm.add_hookspecs(Hook)

    assert (
        repr(pm.hook.myhook.spec)
        == "<HookSpecification 'myhook' args=('arg',)>"
    )

    plugin = Plugin()
    pname = pm.register(plugin)
    expected = f"<HookImplementation plugin={pname!r} spec='myhook'>"
    assert repr(pm.hook.myhook._nonwrappers[0]) == expected



import inspect
import sys
from functools import lru_cache
from typing import Any, Dict, Optional, overload

if sys.version_info >= (3, 8):
    from importlib import metadata as importlib_metadata
else:
    import importlib_metadata


@lru_cache(maxsize=1)
def _top_level_module_to_dist() -> Dict[str, importlib_metadata.Distribution]:
    mapping = {}
    for dist in importlib_metadata.distributions():
        modules = dist.read_text('top_level.txt')
        if modules:
            for mod in modules.split('\n'):
                if not mod:
                    continue
                mapping[mod] = dist
    return mapping


def _object_to_top_level_module(obj: Any) -> Optional[str]:
    module = inspect.getmodule(obj)
    name = getattr(module, '__name__', None)
    return name.split('.')[0] if name else None


def get_dist(obj) -> Optional[importlib_metadata.Distribution]:
    """Return a :class:`importlib.metadata.Distribution` for any python object.

    Parameters
    ----------
    obj : Any
        A python object. If a string, will be interpreted as a dist name.

    Returns
    -------
    dist: Distribution
        The distribution object for the corresponding package, if found.
    """
    if isinstance(obj, str):
        try:
            return importlib_metadata.distribution(obj)
        except importlib_metadata.PackageNotFoundError:
            return None
    top_level = _object_to_top_level_module(obj)
    return _top_level_module_to_dist().get(top_level or '')


def get_version(plugin) -> str:
    dist = get_dist(plugin)
    version = dist.metadata.get('version') if dist else ''
    if not version and inspect.ismodule(plugin):
        version = getattr(plugin, '__version__', '')
    if not version:
        top_module = _object_to_top_level_module(plugin)
        if top_module in sys.modules:
            version = getattr(sys.modules[top_module], '__version__', '')
    return str(version) if version else ''


@overload
def get_metadata(plugin, arg: str, *args: None) -> Optional[str]: ...


@overload  # noqa: F811
def get_metadata(  # noqa: F811
    plugin, arg: str, *args: str
) -> Dict[str, Optional[str]]: ...


def get_metadata(plugin, *args):  # noqa: F811
    """Get metadata for this plugin.

    Valid arguments are any keys from the Core metadata specifications:
    https://packaging.python.org/specifications/core-metadata/

    Parameters
    ----------
    *args : str
        (Case insensitive) names of metadata entries to retrieve.

    Returns
    -------
    str or dict, optional
        If a single argument is provided, the value for that entry is
        returned (which may be ``None``).
        If multiple arguments are provided, a dict of {arg: value} is
        returned.
    """
    dist = get_dist(plugin)
    dct = {}
    if dist:
        for a in args:
            if a == 'version':
                dct[a] = get_version(plugin)
            else:
                dct[a] = dist.metadata.get(a)
    if len(args) == 1:
        return dct[args[0]] if dct else None
    return dct


def standard_metadata(plugin: Any) -> Dict[str, Optional[str]]:
    """Return a standard metadata dict for ``plugin``.

    Parameters
    ----------
    plugin : Any
        A python object.  If a string, will be interpreted as a dist name.

    Returns
    -------
    metadata : dict
        A  dicts with plugin object metadata. The dict is guaranteed to have
        the following keys:

        - **package**: The name of the package
        - **version**: The version of the plugin package
        - **summary**: A one-line summary of what the distribution does
        - **author**: The author’s name
        - **email**: The author’s (or maintainer's) e-mail address.
        - **license**: The license covering the distribution
        - **url**: The home page for the package, or dowload url if N/A.

    Raises
    ------
    ValueError
        If no distribution can be found for ``plugin``.
    """
    meta = {}
    if not get_dist(plugin):
        _top_level_module_to_dist.cache_clear()
        if not get_dist(plugin):
            return {}
            # TODO: decide appropriate behavior here.
            # raise ValueError(f"could not find metadata for {plugin}")
    meta = get_metadata(
        plugin,
        'name',
        'version',
        'summary',
        'author',
        'license',
        'Author-Email',
        'Home-page',
    )
    meta['package'] = meta.pop('name')
    meta['email'] = meta.pop('Author-Email') or get_metadata(
        plugin, 'Maintainer-Email'
    )
    meta['url'] = meta.pop('Home-page') or get_metadata(plugin, 'Download-Url')
    if meta['url'] == 'UNKNOWN':
        meta['url'] = None
    return meta



def test_spec(test_plugin_manager, add_specification):
    assert not test_plugin_manager.hook.items()
    assert not hasattr(test_plugin_manager.hook, 'my_spec')

    @add_specification
    def my_spec(arg1, arg2): ...

    assert hasattr(test_plugin_manager.hook, 'my_spec')
    assert hasattr(test_plugin_manager.hook.my_spec, 'spec')


def test_full(test_plugin_manager, add_specification, add_implementation):
    relay = test_plugin_manager.hook
    assert not relay.items()

    @add_specification
    def my_spec(arg1, arg2): ...

    assert hasattr(relay, 'my_spec')
    assert not relay.my_spec.get_hookimpls()

    @add_implementation(specname='my_spec')
    def ttt(arg):
        return arg + 1

    assert relay.my_spec.get_hookimpls()[0].specname == 'my_spec'


def test_caller_from_implementation(caller_from_implementation):
    def test1(arg):
        return arg + 1

    caller = caller_from_implementation(test1)
    assert caller.get_hookimpls()[0].function == test1
    assert caller(arg=1) == [2]

    def test2(arg):
        return arg + 1

    caller = caller_from_implementation(test2, {'firstresult': True})
    assert caller.get_hookimpls()[0].function == test2
    assert caller(arg=1) == 2



"""Hook annotation decorators"""

from typing import Callable, Optional

from .implementation import HookImplementation, HookSpecification


class HookSpecificationMarker:
    """Decorator helper class for marking functions as hook specifications.

    You can instantiate it with a project_name to get a decorator. Calling
    :py:meth:`.PluginManager.add_hookspecs` later will discover all marked
    functions if the :py:class:`.PluginManager` uses the same project_name.
    """

    def __init__(self, project_name):
        self.project_name = project_name

    def __call__(
        self,
        function: Optional[Callable] = None,
        firstresult: bool = False,
        historic: bool = False,
        warn_on_impl=None,
    ):
        """if passed a function, directly sets attributes on the function
        which will make it discoverable to
        :py:meth:`.PluginManager.add_hookspecs`. If passed no function, returns
        a decorator which can be applied to a function later using the
        attributes supplied.

        If ``firstresult`` is ``True`` the 1:N hook call (N being the number of
        registered hook implementation functions) will stop at I<=N when the
        I'th function returns a non-``None`` result.

        If ``historic`` is ``True`` calls to a hook will be memorized and
        replayed on later registered plugins.

        """

        def setattr_hookspec_opts(func):
            if historic and firstresult:
                raise ValueError("cannot have a historic firstresult hook")
            setattr(
                func,
                HookSpecification.format_tag(self.project_name),
                dict(
                    firstresult=firstresult,
                    historic=historic,
                    warn_on_impl=warn_on_impl,
                ),
            )
            return func

        if function is not None:
            return setattr_hookspec_opts(function)
        else:
            return setattr_hookspec_opts


class HookImplementationMarker:
    """Decorator helper class for marking functions as hook implementations.

    You can instantiate with a ``project_name`` to get a decorator. Calling
    :meth:`.PluginManager.register` later will discover all marked functions if
    the :class:`.PluginManager` uses the same project_name.

    Parameters
    ----------
    project_name : str
        A namespace for plugin implementations.  Implementations decorated with
        this class will be discovered by ``PluginManager.register`` if and only
        if ``project_name`` matches the ``project_name`` of the
        ``PluginManager``.
    """

    def __init__(self, project_name: str):
        self.project_name = project_name

    def __call__(
        self,
        function: Optional[Callable] = None,
        *,
        hookwrapper: bool = False,
        optionalhook: bool = False,
        tryfirst: bool = False,
        trylast: bool = False,
        specname: str = '',
    ) -> Callable:
        """Call the marker instance.

        If passed a function, directly sets attributes on the function which
        will make it discoverable to :meth:`.PluginManager.register`. If passed
        no function, returns a decorator which can be applied to a function
        later using the attributes supplied.

        Parameters
        ----------
        function : callable, optional
            A function to decorate as a hook implementation, If ``function`` is
            None, this method returns a function that can be used to decorate
            other functions.
        hookwrapper : bool, optional
            Whether this hook implementation behaves as a hookwrapper.
            by default False
        optionalhook : bool, optional
            If ``True``, a missing matching hook specification will not result
            in an error (by default it is an error if no matching spec is
            found), by default False.
        tryfirst : bool, optional
            If ``True`` this hook implementation will run as early as possible
            in the chain of N hook implementations for a specification, by
            default False
        trylast : bool, optional
            If ``True`` this hook implementation will run as late as possible
            in the chain of N hook implementations, by default False
        specname : str, optional
            If provided, ``specname`` will be used instead of the function name
            when matching this hook implementation to a hook specification
            during registration, by default the implementation function name
            must match the name of the corresponding hook specification.

        Returns
        -------
        Callable
            If ``function`` is not ``None``, will decorate the function with
            attributes, and return the function.  If ``function`` is None, will
            return a decorator that can be used to decorate functions.
        """

        def set_hook_implementation_attributes(func):
            setattr(
                func,
                HookImplementation.format_tag(self.project_name),
                dict(
                    hookwrapper=hookwrapper,
                    optionalhook=optionalhook,
                    tryfirst=tryfirst,
                    trylast=trylast,
                    specname=specname,
                ),
            )
            return func

        if function is None:
            return set_hook_implementation_attributes
        else:
            return set_hook_implementation_attributes(function)



import pytest

from napari_plugin_engine import (
    HookCallError,
    HookImplementation,
    HookImplementationMarker,
    HookSpecificationMarker,
)
from napari_plugin_engine.callers import _multicall

hookspec = HookSpecificationMarker("example")
example_implementation = HookImplementationMarker("example")


def multicall(methods, kwargs, firstresult=False):
    """utility function to execute the hook implementations loop"""
    caller = _multicall
    hookfuncs = []
    for method in methods:
        f = HookImplementation(method, **method.example_impl)
        hookfuncs.append(f)
    # our _multicall function returns our own HookResult object.
    # so to make these pluggy tests pass, we have to access .result to mimic
    # the old behavior (that directly returns results).
    return caller(hookfuncs, kwargs, firstresult=firstresult).result


def test_multicall_passing():
    class Plugin1:
        @example_implementation
        def method(self, x):
            return 17

    class Plugin2:
        @example_implementation
        def method(self, x):
            return 23

    p1 = Plugin1()
    p2 = Plugin2()
    result_list = multicall([p1.method, p2.method], {"x": 23})
    assert len(result_list) == 2
    # ensure reversed order
    assert result_list == [23, 17]


def test_keyword_args():
    @example_implementation
    def func(x):
        return x + 1

    class Plugin:
        @example_implementation
        def func(self, x, y):
            return x + y

    reslist = multicall([func, Plugin().func], {"x": 23, "y": 24})
    assert reslist == [24 + 23, 24]


def test_keyword_args_with_defaultargs():
    @example_implementation
    def func(x, z=1):
        return x + z

    reslist = multicall([func], {"x": 23, "y": 24})
    assert reslist == [24]


def test_tags_call_error():
    @example_implementation
    def func(x):
        return x

    with pytest.raises(HookCallError):
        multicall([func], {})


def test_call_subexecute():
    @example_implementation
    def func1():
        return 2

    @example_implementation
    def func2():
        return 1

    assert multicall([func2, func1], {}, firstresult=True) == 2


def test_call_none_is_no_result():
    @example_implementation
    def func1():
        return 1

    @example_implementation
    def func2():
        return None

    assert multicall([func1, func2], {}, firstresult=True) == 1
    assert multicall([func1, func2], {}, {}) == [1]


def test_hookwrapper():
    out = []

    @example_implementation(hookwrapper=True)
    def func1():
        out.append("func1 init")
        yield None
        out.append("func1 finish")

    @example_implementation
    def func2():
        out.append("func2")
        return 2

    assert multicall([func2, func1], {}) == [2]
    assert out == ["func1 init", "func2", "func1 finish"]
    out = []
    assert multicall([func2, func1], {}, firstresult=True) == 2
    assert out == ["func1 init", "func2", "func1 finish"]


def test_hookwrapper_order():
    out = []

    @example_implementation(hookwrapper=True)
    def func1():
        out.append("func1 init")
        yield 1
        out.append("func1 finish")

    @example_implementation(hookwrapper=True)
    def func2():
        out.append("func2 init")
        yield 2
        out.append("func2 finish")

    assert multicall([func2, func1], {}) == []
    assert out == ["func1 init", "func2 init", "func2 finish", "func1 finish"]


def test_hookwrapper_not_yield():
    @example_implementation(hookwrapper=True)
    def func1():
        pass

    with pytest.raises(TypeError):
        multicall([func1], {})


def test_hookwrapper_too_many_yield():
    @example_implementation(hookwrapper=True)
    def func1():
        yield 1
        yield 2

    with pytest.raises(RuntimeError) as ex:
        multicall([func1], {})
    assert "func1" in str(ex.value)
    assert __file__ in str(ex.value)


@pytest.mark.parametrize("exc", [SystemExit])
def test_hookwrapper_exception(exc):
    out = []

    @example_implementation(hookwrapper=True)
    def func1():
        out.append("func1 init")
        yield None
        out.append("func1 finish")

    @example_implementation
    def func2():
        raise exc()

    with pytest.raises(exc):
        multicall([func2, func1], {})
    assert out == ["func1 init", "func1 finish"]



import inspect
import sys
from typing import Any, Callable, Optional


class HookImplementation:
    """A class to encapsulate hook implementations."""

    TAG_SUFFIX = "_impl"

    def __init__(
        self,
        function: Callable,
        plugin: Optional[Any] = None,
        plugin_name: Optional[str] = None,
        hookwrapper: bool = False,
        optionalhook: bool = False,
        tryfirst: bool = False,
        trylast: bool = False,
        specname: str = '',
        enabled: bool = True,
    ):
        self.function = function
        self.argnames, self.kwargnames = varnames(self.function)
        self.plugin = plugin
        self.plugin_name = plugin_name
        self.hookwrapper = hookwrapper
        self.optionalhook = optionalhook
        self.tryfirst = tryfirst
        self.trylast = trylast
        self._specname = specname
        self.enabled = enabled

    @classmethod
    def format_tag(cls, project_name):
        return project_name + cls.TAG_SUFFIX

    @property
    def opts(self) -> dict:
        # legacy
        return {
            x: getattr(self, x)
            for x in [
                'hookwrapper',
                'optionalhook',
                'tryfirst',
                'trylast',
                'specname',
            ]
        }

    def __repr__(self) -> str:
        # these are all False by default
        truthy = [
            attr
            for attr in ('hookwrapper', 'optionalhook', 'tryfirst', 'trylast')
            if getattr(self, attr)
        ]
        suffix = (' ' + " ".join(truthy)) if truthy else ''
        return (
            f"<HookImplementation plugin={self.plugin_name!r}"
            f" spec={self.specname!r}{suffix}>"
        )

    def __call__(self, *args):
        return self.function(*args)

    @property
    def specname(self) -> str:
        return self._specname or self.function.__name__


class HookSpecification:
    """A class to encapsulate hook specifications."""

    TAG_SUFFIX = "_spec"

    def __init__(
        self,
        namespace: Any,
        name: str,
        *,
        firstresult: bool = False,
        historic: bool = False,
        warn_on_impl: Optional[Warning] = None,
    ):
        self.namespace = namespace
        self.name = name
        self.function = getattr(namespace, name)
        self.argnames, self.kwargnames = varnames(self.function)
        for reserved in ('_plugin', '_skip_impls'):
            if reserved in self.argnames:
                raise ValueError(
                    f'Hook specifications may not have argument: "{reserved}".'
                )
        self.firstresult = firstresult
        self.historic = historic
        self.warn_on_impl = warn_on_impl

    @classmethod
    def format_tag(cls, project_name):
        return project_name + cls.TAG_SUFFIX

    @property
    def opts(self) -> dict:
        # legacy
        return {
            'firstresult': self.firstresult,
            'historic': self.historic,
            'warn_on_impl': self.warn_on_impl,
        }

    def __repr__(self) -> str:
        # these are all False by default
        truthy = [
            attr
            for attr in ('firstresult', 'historic', 'warn_on_impl')
            if getattr(self, attr)
        ]
        suffix = (' ' + " ".join(truthy)) if truthy else ''
        return (
            f"<HookSpecification {self.name!r} args={self.argnames!r}{suffix}>"
        )


# TODO: can this be improved?
def varnames(func):
    """Return tuple of positional and keywrord argument names for a function,
    method, class or callable.

    In case of a class, its ``__init__`` method is considered.
    For methods the ``self`` parameter is not included.
    """
    cache = getattr(func, "__dict__", {})
    try:
        return cache["_varnames"]
    except KeyError:
        pass

    if inspect.isclass(func):
        try:
            func = func.__init__
        except AttributeError:
            return (), ()
    elif not inspect.isroutine(func):  # callable object?
        try:
            func = getattr(func, "__call__", func)
        except Exception:
            return (), ()

    try:  # func MUST be a function or method here or we won't parse any args
        spec = inspect.getfullargspec(func)
    except TypeError:
        return (), ()

    args, defaults = tuple(spec.args), spec.defaults
    if defaults:
        index = -len(defaults)
        args, kwargs = args[:index], tuple(args[index:])
    else:
        kwargs = ()

    # strip any implicit instance arg
    # pypy3 uses "obj" instead of "self" for default dunder methods
    _PYPY3 = hasattr(sys, "pypy_version_info") and sys.version_info.major == 3
    implicit_names = ("self",) if not _PYPY3 else ("self", "obj")
    if args:
        if inspect.ismethod(func) or (
            "." in getattr(func, "__qualname__", ())
            and args[0] in implicit_names
        ):
            args = args[1:]

    try:
        cache["_varnames"] = args, kwargs
    except TypeError:
        pass
    return args, kwargs
